# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Represent a ``PlanFactory`` for tracking workflows."""

import os
import re
import shlex
import time
from contextlib import contextmanager
from datetime import datetime
from itertools import chain
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Set, Tuple, Union, cast

import click
import yaml

from renku.command.command_builder.command import inject
from renku.core import errors
from renku.core.constant import RENKU_HOME, RENKU_TMP
from renku.core.interface.project_gateway import IProjectGateway
from renku.core.plugin.pluginmanager import get_plugin_manager
from renku.core.storage import check_external_storage, track_paths_in_storage
from renku.core.util.git import is_path_safe
from renku.core.util.metadata import is_external_file
from renku.core.util.os import get_absolute_path, get_relative_path, is_subpath
from renku.core.workflow.types import PATH_OBJECTS, Directory, File
from renku.domain_model.datastructures import DirectoryTree
from renku.domain_model.project_context import project_context
from renku.domain_model.provenance.agent import Person
from renku.domain_model.workflow.parameter import (
    DIRECTORY_MIME_TYPE,
    CommandInput,
    CommandOutput,
    CommandParameter,
    MappedIOStream,
)
from renku.domain_model.workflow.plan import Plan

STARTED_AT = int(time.time() * 1000)


class PlanFactory:
    """Factory for creating a plan from a command line call."""

    _RE_SUBCOMMAND = re.compile(r"^[A-Za-z]+(-[A-Za-z]+)?$")

    command_line: List[str]

    def __init__(
        self,
        command_line: Union[str, List[str], Tuple[str, ...]],
        explicit_inputs: Optional[List[Tuple[str, str]]] = None,
        explicit_outputs: Optional[List[Tuple[str, str]]] = None,
        explicit_parameters: Optional[List[Tuple[str, Optional[str]]]] = None,
        directory: Optional[Union[Path, str]] = None,
        working_dir: Optional[Union[Path, str]] = None,
        no_input_detection: bool = False,
        no_output_detection: bool = False,
        no_parameter_detection: bool = False,
        success_codes: Optional[List[int]] = None,
        stdin: Optional[str] = None,
        stdout: Optional[str] = None,
        stderr: Optional[str] = None,
    ):
        self.plan_id = Plan.generate_id()

        self.no_input_detection = no_input_detection
        self.no_output_detection = no_output_detection
        self.no_parameter_detection = no_parameter_detection

        if not command_line:
            raise errors.UsageError("Command line can not be empty. Please specify a command to execute.")

        if not directory:
            directory = os.getcwd()
        self.directory = Path(directory)
        if not self.directory.exists():
            raise errors.UsageError(
                f"Directory '{self.directory}' doesn't exist. Please make sure you are inside an existing directory."
            )

        if not working_dir:
            working_dir = os.getcwd()
        self.working_dir = Path(working_dir)
        if not self.working_dir.exists():
            raise errors.UsageError(
                f"Repository path '{self.working_dir}' doesn't exist. Make sure you are inside a Renku repository."
            )

        if isinstance(command_line, (list, tuple)):
            self.command_line = list(command_line)
        else:
            self.command_line = shlex.split(command_line)

        self.success_codes = success_codes or []

        self.explicit_inputs: List[Tuple[str, str]] = (
            [(get_absolute_path(path), name) for path, name in explicit_inputs] if explicit_inputs else []
        )
        self.explicit_outputs: List[Tuple[str, str]] = (
            [(get_absolute_path(path), name) for path, name in explicit_outputs] if explicit_outputs else []
        )
        self.explicit_parameters = explicit_parameters if explicit_parameters else []

        self.stdin = stdin
        self.stdout = stdout
        self.stderr = stderr

        self.base_command, detected_arguments = self.split_command_and_args()
        self.parameters: List[CommandParameter] = []
        self.inputs: List[CommandInput] = []
        self.outputs: List[CommandOutput] = []
        self.annotations: List[Dict[str, Any]] = []
        self.existing_directories: Set[str] = set()

        self.add_inputs_and_parameters(detected_arguments)

    def split_command_and_args(self):
        """Return tuple with command and args from command line arguments."""
        existing_subpath = self._resolve_existing_subpath(self.command_line[0])
        if existing_subpath is not None:
            return [], list(self.command_line)

        cmd = [self.command_line[0]]
        args = list(self.command_line[1:])

        if len(args) < 2:
            # only guess subcommand for more arguments
            return cmd, args

        while args and re.match(self._RE_SUBCOMMAND, args[0]) and not self._resolve_existing_subpath(args[0]):
            cmd.append(args.pop(0))

        return cmd, args

    @staticmethod
    def _is_ignored_path(candidate: Union[Path, str], ignored_list: Optional[Set[str]] = None) -> bool:
        """Return True if the path is in ignored list."""
        return ignored_list is not None and str(candidate) in ignored_list

    def _resolve_existing_subpath(self, candidate: Union[Path, str]) -> Optional[Path]:
        """Return a path instance if it exists in the project's directory."""
        candidate = self.directory / candidate if not os.path.isabs(candidate) else Path(candidate)

        if candidate.exists() or candidate.is_symlink():
            path = candidate.resolve()

            # NOTE: If resolved path is not within the project then it's is either an external file or an absolute path
            # (e.g. /bin/bash)
            if is_subpath(path, base=self.working_dir):
                return path
            elif is_external_file(path=candidate, project_path=self.working_dir):
                return Path(os.path.abspath(candidate))

        return None

    def add_inputs_and_parameters(self, arguments: List[str]):
        """Yield command input parameters."""
        position = 0
        prefix: Optional[str] = None

        output_streams = {getattr(self, stream_name) for stream_name in ("stdout", "stderr")}

        for index, argument in enumerate(arguments):
            if prefix:
                if argument.startswith("-"):
                    position += 1
                    self.add_command_parameter(default_value=prefix, position=position)
                    prefix = None

            if argument.startswith("--"):
                if "=" in argument:
                    prefix, value = argument.split("=", 1)
                    prefix += "="
                    default, type = self.guess_type(value, ignore_filenames=output_streams)

                    position += 1
                    if type in PATH_OBJECTS:
                        self.add_command_input(
                            default_value=self._path_relative_to_root(default.path),
                            prefix=prefix,
                            position=position,
                            encoding_format=[DIRECTORY_MIME_TYPE] if type == "Directory" else default.mime_type,
                        )
                    else:
                        self.add_command_parameter(default_value=default, prefix=prefix, position=position)

                    prefix = None
                else:
                    prefix = argument

            elif argument.startswith("-"):
                if len(argument) > 2:
                    if "=" in argument:
                        prefix, value = argument.split("=", 1)
                        prefix += "="
                        default, type = self.guess_type(value, ignore_filenames=output_streams)
                    else:
                        # possibly a flag with value
                        prefix = argument[0:2]
                        default, type = self.guess_type(argument[2:], ignore_filenames=output_streams)

                    position += 1

                    if type in PATH_OBJECTS:
                        self.add_command_input(
                            default_value=self._path_relative_to_root(default.path),
                            prefix=prefix,
                            position=position,
                            encoding_format=[DIRECTORY_MIME_TYPE] if type == "Directory" else default.mime_type,
                        )
                    else:
                        self.add_command_parameter(default_value=default, prefix=prefix, position=position)
                    prefix = None
                else:
                    prefix = argument

            else:
                default, type = self.guess_type(argument, ignore_filenames=output_streams)

                position += 1

                if prefix:
                    prefix = prefix + " "  # NOTE: Make sure prefix is separated from argument by space

                if type in PATH_OBJECTS:
                    self.add_command_input(
                        default_value=self._path_relative_to_root(default.path),
                        prefix=prefix,
                        position=position,
                        encoding_format=[DIRECTORY_MIME_TYPE] if type == "Directory" else default.mime_type,
                    )
                else:
                    self.add_command_parameter(default_value=default, prefix=prefix, position=position)
                prefix = None

        if prefix:
            position += 1
            self.add_command_parameter(default_value=prefix, position=position)

        if self.stdin:
            position += 1
            default, type = self.guess_type(str(self.working_dir / self.stdin), ignore_filenames=output_streams)
            assert isinstance(default, File)
            self.add_command_input(default_value=str(default), encoding_format=default.mime_type, position=position)

    def add_outputs(self, candidates: Iterable[Tuple[Union[Path, str], Optional[str]]]):
        """Yield detected output and changed command input parameter."""
        # TODO what to do with duplicate paths & inputs with same defaults
        candidate_paths = list(map(lambda x: x[0], candidates))
        tree = DirectoryTree.from_list(candidate_paths)

        input_candidates = {}
        parameter_candidates = {}

        for input in self.inputs:
            # NOTE: Check if an input directory was modified
            input_path = Path(input.default_value)

            if input_path.is_dir() and tree.get(input_path):
                # The directory might exist before running the script
                candidates = self._check_potential_output_directory(input_path, candidates, tree)  # type: ignore

            input_candidates[str(input_path)] = input

        for parameter in self.parameters:
            # NOTE: find parameters that might actually be outputs
            if any(parameter.default_value == value for value, _ in self.explicit_parameters):
                continue

            try:
                path = self.directory / str(parameter.default_value)
                input_path = Path(os.path.abspath(path)).relative_to(self.working_dir)
            except FileNotFoundError:
                continue
            except ValueError:
                # NOTE: Raised if path is not relative to working_dir (external file)
                input_path = Path(parameter.default_value)

                if not input_path.exists():
                    continue

            if input_path.is_dir() and tree.get(input_path):
                # The directory might exist before running the script
                candidates = self._check_potential_output_directory(input_path, candidates, tree)  # type: ignore

                parameter_candidates[str(input_path)] = parameter
            parameter_candidates[str(input_path)] = parameter

        for candidate_path, name in candidates:
            candidate = self._resolve_existing_subpath(self.working_dir / candidate_path)

            if candidate is None:
                raise errors.UsageError(f'Path "{candidate_path}" does not exist inside the current project.')

            glob = str(candidate.relative_to(self.working_dir))

            if glob in input_candidates:
                input = input_candidates[glob]

                self.add_command_output_from_input(input, name=name)
            elif glob in parameter_candidates:
                param = parameter_candidates[glob]

                self.add_command_output_from_parameter(param, name=name)
            else:
                encoding_format = [DIRECTORY_MIME_TYPE] if candidate.is_dir() else self._get_mimetype(candidate)
                self.add_command_output(default_value=glob, encoding_format=encoding_format, name=name)

    def _check_potential_output_directory(
        self, input_path: Path, candidates: Set[Tuple[str, Optional[str]]], tree: DirectoryTree
    ) -> Set[Tuple[str, Optional[str]]]:
        """Check an input/parameter for being a potential output directory."""
        subpaths = {str(input_path / path) for path in tree.get(input_path, default=[])}
        if not self._is_explicit(input_path, self.explicit_outputs):
            content = {str(path) for path in input_path.rglob("*") if not path.is_dir() and path.name != ".gitkeep"}
            preexisting_paths = content - subpaths
            if preexisting_paths:
                raise errors.InvalidOutputPath(
                    f"The output directory '{input_path}' is not empty. \n\n"
                    "As renku treats whole directory outputs as generated by renku,"
                    "those directories have to be empty before being tracked by renku.\n\n"
                    "You can solve this by:\n"
                    f"- Deleting the existing files in the directory (use 'git rm -r {input_path}')\n"
                    "- Using a different output folder\n"
                    "- Using a new empty subfolder inside the output folder\n\n"
                    "Output directories with existing files:\n\n"
                    + "\n".join("\t" + click.style(path, fg="yellow") for path in preexisting_paths)
                    + "\n\n"
                    "Once you have resolved the issues above, you can safely rerun the previous command."
                )

        # Remove files from the input directory
        candidates = {(path, name) for path, name in candidates if path not in subpaths}
        # Include input path in the candidates to check
        candidates.add((str(input_path), None))

        return candidates

    @staticmethod
    def _get_mimetype(file: Path) -> List[str]:
        """Return the MIME-TYPE of the given file."""
        # TODO: specify the actual mime-type of the file
        return ["application/octet-stream"]

    def guess_type(self, value: Union[Path, str], ignore_filenames: Optional[Set[str]] = None) -> Tuple[Any, str]:
        """Return new value and CWL parameter type."""
        if not self._is_ignored_path(value, ignore_filenames) and all(value != v for v, _ in self.explicit_parameters):
            candidate = self._resolve_existing_subpath(value)
            if candidate:
                if candidate.is_dir():
                    return Directory(path=candidate), "Directory"
                return File(path=candidate, mime_type=self._get_mimetype(candidate)), "File"

        return value, "string"

    def get_stream_mapping_for_value(self, value: Any):
        """Return a stream mapping if value is a path mapped to a stream."""
        if self.stdin and self.stdin == value:
            return MappedIOStream(id=MappedIOStream.generate_id("stdin"), stream_type="stdin")

        if self.stdout and self.stdout == value:
            return MappedIOStream(id=MappedIOStream.generate_id("stdout"), stream_type="stdout")
        if self.stderr and self.stderr == value:
            return MappedIOStream(id=MappedIOStream.generate_id("stderr"), stream_type="stderr")

    def add_command_input(
        self,
        default_value: Any,
        prefix: Optional[str] = None,
        position: Optional[int] = None,
        postfix: Optional[str] = None,
        name: Optional[str] = None,
        encoding_format: Optional[List[str]] = None,
    ):
        """Create a CommandInput."""
        if self.no_input_detection and not self._is_explicit(default_value, self.explicit_inputs):
            return

        mapped_stream = self.get_stream_mapping_for_value(default_value)

        inp_param = CommandInput(
            id=CommandInput.generate_id(
                plan_id=self.plan_id,
                position=position,
                postfix=mapped_stream.stream_type if mapped_stream else postfix,
            ),
            default_value=default_value,
            prefix=prefix,
            position=position,
            mapped_to=mapped_stream,
            encoding_format=encoding_format,
            postfix=postfix,
            name=name,
        )

        existing_parameter = next((p for p in self.inputs if p.name == inp_param.name), None)

        if existing_parameter is not None and existing_parameter.default_value == inp_param.default_value:
            existing_parameter.update_from(inp_param)
        elif existing_parameter is not None:
            # duplicate with different values!
            raise errors.ParameterError(
                f"Duplicate input '{inp_param.name}' found with differing values ('{inp_param.default_value}'"
                f" vs. '{existing_parameter.default_value}')"
            )
        else:
            self.inputs.append(inp_param)

    def add_command_output(
        self,
        default_value: Any,
        prefix: Optional[str] = None,
        position: Optional[int] = None,
        postfix: Optional[str] = None,
        encoding_format: Optional[List[str]] = None,
        name: Optional[str] = None,
        id: Optional[str] = None,
        mapped_to: Optional[MappedIOStream] = None,
    ):
        """Create a CommandOutput."""
        if self.no_output_detection and not self._is_explicit(default_value, self.explicit_outputs):
            return

        create_folder = False
        path = Path(default_value)
        full_path = Path(self._path_relative_to_root(default_value)).resolve()

        if (full_path.is_dir() and str(path) in self.existing_directories) or (
            not full_path.is_dir() and str(path.parent) in self.existing_directories
        ):
            create_folder = True

        mapped_stream = mapped_to or self.get_stream_mapping_for_value(default_value)

        if mapped_stream and position is None:
            position = (
                max((p.position for p in chain(self.inputs, self.outputs, self.parameters) if p.position), default=0)
                + 1
            )

        id = id or CommandOutput.generate_id(
            plan_id=self.plan_id,
            position=position,
            postfix=mapped_stream.stream_type if mapped_stream else postfix,
        )

        out_param = CommandOutput(
            id=id,
            default_value=default_value,
            prefix=prefix,
            position=position,
            mapped_to=mapped_stream,
            encoding_format=encoding_format,
            postfix=postfix,
            create_folder=create_folder,
            name=name,
        )

        existing_parameter = next((p for p in self.outputs if p.name == out_param.name), None)

        if existing_parameter is not None and existing_parameter.default_value == out_param.default_value:
            existing_parameter.update_from(out_param)
        elif existing_parameter is not None:
            # duplicate with different values!
            raise errors.ParameterError(
                f"Duplicate output '{out_param.name}' found with differing values ('{out_param.default_value}'"
                f" vs. '{existing_parameter.default_value}')"
            )
        else:
            self.outputs.append(out_param)

    def add_command_output_from_input(self, input: CommandInput, name):
        """Create a CommandOutput from an input."""
        self.inputs.remove(input)

        self.add_command_output(
            id=input.id.replace("/inputs/", "/outputs/"),
            default_value=input.default_value,
            prefix=input.prefix,
            position=input.position,
            mapped_to=input.mapped_to,
            encoding_format=input.encoding_format,
            name=name,
        )

    def add_command_output_from_parameter(self, parameter: CommandParameter, name):
        """Create a CommandOutput from a parameter."""
        self.parameters.remove(parameter)
        value = Path(self._path_relative_to_root(parameter.default_value))
        encoding_format = [DIRECTORY_MIME_TYPE] if value.is_dir() else self._get_mimetype(value)
        self.add_command_output(
            default_value=str(value),
            prefix=parameter.prefix,
            position=parameter.position,
            encoding_format=encoding_format,
            name=name,
        )

    def add_command_parameter(
        self,
        default_value: Any,
        prefix: Optional[str] = None,
        position: Optional[int] = None,
        name: Optional[str] = None,
    ):
        """Create a CommandParameter."""
        if self.no_parameter_detection and all(default_value != v for v, _ in self.explicit_parameters):
            return

        parameter = CommandParameter(
            id=CommandParameter.generate_id(plan_id=self.plan_id, position=position),
            default_value=default_value,
            prefix=prefix,
            position=position,
            name=name,
        )

        existing_parameter = next((p for p in self.parameters if p.name == parameter.name), None)

        if existing_parameter is not None and existing_parameter.default_value == parameter.default_value:
            existing_parameter.update_from(parameter)
        elif existing_parameter is not None:
            # duplicate with different values!
            raise errors.ParameterError(
                f"Duplicate parameter '{parameter.name}' found with differing values ('{parameter.default_value}'"
                f" vs. '{existing_parameter.default_value}')"
            )
        else:
            self.parameters.append(parameter)

    def add_explicit_inputs(self):
        """Add explicit inputs ."""
        input_paths = [input.default_value for input in self.inputs]
        input_id = len(self.inputs) + len(self.parameters)

        for explicit_input, name in self.explicit_inputs:
            try:
                relative_explicit_input = get_relative_path(explicit_input, base=self.working_dir, strict=True)
            except errors.ParameterError:
                raise errors.UsageError(
                    "The input file or directory is not in the repository."
                    "\n\n\t" + click.style(str(explicit_input), fg="yellow") + "\n\n"
                )

            if relative_explicit_input in input_paths:
                if name:
                    existing_inputs = [i for i in self.inputs if i.default_value == relative_explicit_input]

                    for existing_input in existing_inputs:
                        existing_input.name = name
                continue

            input_paths.append(explicit_input)

            if self._resolve_existing_subpath(explicit_input) is None:
                raise errors.UsageError(
                    "The input file or directory does not exist."
                    "\n\n\t" + click.style(str(explicit_input), fg="yellow") + "\n\n"
                )
            input_id += 1
            default, type = self.guess_type(explicit_input)
            # Explicit inputs are either File or Directory
            assert type in PATH_OBJECTS
            self.add_command_input(
                default_value=str(default),
                postfix=str(input_id),
                encoding_format=[DIRECTORY_MIME_TYPE] if type == "Directory" else default.mime_type,
                name=name,
            )

    def add_explicit_parameters(self):
        """Add explicit parameters."""
        parameter_names = [parameter.name for parameter in self.parameters]

        for explicit_parameter, name in self.explicit_parameters:
            if name and name in parameter_names:
                continue

            existing_parameters = [p for p in self.parameters if p.default_value == explicit_parameter]

            if existing_parameters and name:
                # NOTE: Update names of existing parameters
                for p in existing_parameters:
                    p.name = name
            elif not existing_parameters:
                self.add_command_parameter(explicit_parameter, name=name)

    @contextmanager
    def watch(self, no_output=False):
        """Watch a Renku repository for changes to detect outputs."""
        check_external_storage()

        repository = project_context.repository

        # Remove indirect files list if any
        delete_indirect_files_list(self.working_dir)

        from renku.core.plugin.pluginmanager import get_plugin_manager

        pm = get_plugin_manager()
        pm.hook.pre_run(tool=self)
        self.existing_directories = {str(p.relative_to(project_context.path)) for p in project_context.path.glob("**/")}

        yield self

        if repository:
            # Include indirect inputs and outputs before further processing
            self.add_indirect_inputs()
            self.add_indirect_outputs()

            self._include_indirect_parameters()

            # Remove indirect files list if any
            delete_indirect_files_list(self.working_dir)

            # List of all output paths.
            output_paths = []

            # Keep track of unmodified output files.
            unmodified = set()

            candidates: Set[Tuple[Union[Path, str], Optional[str]]] = set()

            if not self.no_output_detection:
                # Calculate possible output paths.
                # Capture newly created files through redirects.
                candidates |= {(file_, None) for file_ in repository.untracked_files}

                # Capture modified files through redirects.
                candidates |= {(o.b_path, None) for o in repository.unstaged_changes if not o.deleted}

                # Filter out explicit outputs
                explicit_output_paths = {
                    str(Path(path).relative_to(self.working_dir)) for path, _ in self.explicit_outputs
                }
                candidates = {c for c in candidates if c[0] not in explicit_output_paths}

            # Include explicit outputs
            candidates |= {
                (str(Path(path).relative_to(self.working_dir)), name) for path, name in self.explicit_outputs
            }

            candidates = {(path, name) for path, name in candidates if is_path_safe(path)}

            self.add_outputs(candidates)

            for stream_name in ("stdout", "stderr"):
                stream = getattr(self, stream_name)
                if (
                    stream
                    and all(stream != path for path, _ in candidates)
                    and not self._is_explicit(stream, self.explicit_outputs)
                ):
                    unmodified.add(stream)
                elif stream:
                    output_paths.append(stream)

            for output in self.outputs:
                if output.default_value not in output_paths:
                    output_paths.append(output.default_value)

            if unmodified:
                raise errors.UnmodifiedOutputs(repository, unmodified)

            if not no_output and not output_paths:
                raise errors.OutputsNotFound()

            if check_external_storage():
                track_paths_in_storage(*output_paths)

            repository.add(*output_paths)

    def _path_relative_to_root(self, path) -> str:
        """Make a potentially relative path in a subdirectory relative to the root of the repository."""
        absolute_path = get_absolute_path(path, base=self.directory)
        return cast(str, get_relative_path(absolute_path, base=self.working_dir, strict=True))

    def _include_indirect_parameters(self):
        run_parameters = read_indirect_parameters(self.working_dir)

        for k, v in run_parameters.items():
            self.explicit_parameters.append((str(v), k))

        self.add_explicit_parameters()

    def add_indirect_inputs(self):
        """Read indirect inputs list and add them to explicit inputs."""
        indirect_inputs_list = get_indirect_inputs_path(self.working_dir)

        for name, indirect_input in read_files_list(indirect_inputs_list).items():
            # treat indirect inputs like explicit inputs
            path = get_absolute_path(indirect_input)
            self.explicit_inputs.append((path, name))

        # add new explicit inputs (if any) to inputs
        self.add_explicit_inputs()

    def add_indirect_outputs(self):
        """Read indirect outputs list and add them to explicit outputs."""
        indirect_outputs_list = get_indirect_outputs_path(self.working_dir)

        for name, indirect_output in read_files_list(indirect_outputs_list).items():
            # treat indirect outputs like explicit outputs
            path = get_absolute_path(indirect_output)
            self.explicit_outputs.append((path, name))

    def iter_input_files(self, basedir):
        """Yield tuples with input id and path."""
        for input_ in self.inputs:
            yield input_.id, os.path.normpath(os.path.join(basedir, input_.default_value))

    @staticmethod
    def _is_explicit(path: Union[Path, str], explicits_collection: List[Tuple[str, str]]) -> bool:
        absolute_path = get_absolute_path(path)
        return any(absolute_path == path for path, _ in explicits_collection)

    @inject.autoparams("project_gateway")
    def to_plan(
        self,
        project_gateway: IProjectGateway,
        name: Optional[str] = None,
        description: Optional[str] = None,
        keywords: Optional[List[str]] = None,
        creators: Optional[List[Person]] = None,
        date_created: Optional[datetime] = None,
    ) -> Plan:
        """Return an instance of ``Plan`` based on this factory."""
        plan = Plan(
            id=self.plan_id,
            name=name,
            date_created=date_created,
            description=description,
            keywords=keywords,
            command=" ".join(self.base_command),
            inputs=self.inputs,
            outputs=self.outputs,
            parameters=self.parameters,
            project_id=project_gateway.get_project().id,
            success_codes=self.success_codes,
            creators=creators,
        )

        pm = get_plugin_manager()

        plugin_annotations = list(chain.from_iterable(pm.hook.plan_annotations(plan=plan)))

        if plugin_annotations:
            plan.annotations.extend(plugin_annotations)

        return plan


def read_files_list(files_list: Path):
    """Read files list yaml containing name:path pairs."""

    if not files_list.exists():
        return {}

    data = yaml.safe_load(files_list.read_text())

    if not isinstance(data, dict):
        raise errors.OperationError("Explicit Inputs/Outputs/Parameters files list must be a YAML dictionary.")

    return data


def add_to_files_list(file_list_path: Path, name: str, path: Union[str, Path]):
    """Add a parameter to indirect parameters."""
    data = read_files_list(file_list_path)

    if name in data and data[name] != str(path):
        raise errors.ParameterError(f"Duplicate input/output name found: {name}")

    data[name] = str(path)

    yaml_data = yaml.dump(data)

    file_list_path.parent.mkdir(exist_ok=True, parents=True)
    file_list_path.write_text(yaml_data)


def delete_indirect_files_list(working_dir):
    """Remove indirect inputs, outputs, and parameters list."""
    paths = [
        get_indirect_inputs_path(working_dir),
        get_indirect_outputs_path(working_dir),
        get_indirect_parameters_path(working_dir),
    ]
    for path in paths:
        try:
            os.remove(path)
        except FileNotFoundError:
            pass


def get_indirect_inputs_path(project_path):
    """Return path to file that contains indirect inputs list."""
    parent = _get_indirect_parent_path(project_path)
    return parent / "inputs.yml"


def get_indirect_outputs_path(project_path):
    """Return path to file that contains indirect outputs list."""
    parent = _get_indirect_parent_path(project_path)
    return parent / "outputs.yml"


def get_indirect_parameters_path(project_path):
    """Return path to file that contains indirect parameters list."""
    parent = _get_indirect_parent_path(project_path)
    return parent / "parameters.yml"


def _get_indirect_parent_path(project_path):
    renku_indirect_path = os.getenv("RENKU_INDIRECT_PATH") or ""

    base = (Path(project_path) / RENKU_HOME / RENKU_TMP).resolve()
    parent = (base / renku_indirect_path).resolve()

    try:
        parent.relative_to(base)
    except ValueError:
        raise errors.InvalidFileOperation(f"Invalid value for RENKU_INDIRECT_PATH env var: {renku_indirect_path}.")

    return parent


def read_indirect_parameters(working_dir):
    """Read and return indirect parameters."""
    path = get_indirect_parameters_path(working_dir)

    if not path.exists():
        return {}

    data = yaml.safe_load(path.read_text())

    if not isinstance(data, dict):
        raise errors.OperationError("Run parameters must be a dictionary.")

    return data


def add_indirect_parameter(working_dir, name, value):
    """Add a parameter to indirect parameters."""
    data = read_indirect_parameters(working_dir)

    if name in data and data[name] != value:
        raise errors.ParameterError(f"Duplicate parameter names found: {name}")
    data[name] = value

    yaml_data = yaml.dump(data)

    path = get_indirect_parameters_path(working_dir)
    path.parent.mkdir(exist_ok=True, parents=True)
    path.write_text(yaml_data)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""S3 dataset provider."""

import urllib
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional, Tuple, cast

from renku.core import errors
from renku.core.dataset.providers.api import (
    AddProviderInterface,
    ProviderApi,
    ProviderCredentials,
    ProviderPriority,
    StorageProviderInterface,
)
from renku.core.dataset.providers.common import get_metadata
from renku.core.dataset.providers.models import DatasetAddAction
from renku.core.interface.storage import IStorage
from renku.core.util.metadata import prompt_for_credentials
from renku.core.util.urls import get_scheme
from renku.domain_model.project_context import project_context
from renku.infrastructure.storage.factory import StorageFactory

if TYPE_CHECKING:
    from renku.core.dataset.providers.models import DatasetAddMetadata
    from renku.domain_model.dataset import Dataset


class S3Provider(ProviderApi, StorageProviderInterface, AddProviderInterface):
    """S3 provider."""

    priority = ProviderPriority.HIGHEST
    name = "S3"
    is_remote = True

    def __init__(self, uri: str):
        super().__init__(uri=uri)

        endpoint, bucket, _ = parse_s3_uri(uri=self.uri)

        self._bucket: str = bucket
        self._endpoint: str = endpoint

    @staticmethod
    def supports(uri: str) -> bool:
        """Whether or not this provider supports a given URI."""
        return get_scheme(uri) == "s3"

    @staticmethod
    def supports_storage(uri: str) -> bool:
        """Whether or not this provider supports a given URI storage."""
        return S3Provider.supports(uri=uri)

    def get_metadata(
        self, uri: str, destination: Path, dataset_add_action: DatasetAddAction = DatasetAddAction.NONE, **_
    ) -> List["DatasetAddMetadata"]:
        """Get metadata of files that will be added to a dataset."""
        return get_metadata(provider=self, uri=uri, destination=destination, dataset_add_action=dataset_add_action)

    def convert_to_storage_uri(self, uri: str) -> str:
        """Convert backend-specific URI to a URI that is usable by the IStorage implementation."""
        _, bucket, path = parse_s3_uri(uri=uri)
        return f"s3://{bucket}/{path}"

    def get_credentials(self) -> "S3Credentials":
        """Return an instance of provider's credential class."""
        return S3Credentials(provider=self)

    def get_storage(self, credentials: Optional["ProviderCredentials"] = None) -> "IStorage":
        """Return the storage manager for the provider."""
        s3_configuration = {
            "type": "s3",
            "provider": "AWS",
            "endpoint": self.endpoint,
        }

        if not credentials:
            credentials = self.get_credentials()
            prompt_for_credentials(credentials)

        return StorageFactory.get_storage(
            storage_scheme="s3",
            provider=self,
            credentials=credentials,
            configuration=s3_configuration,
        )

    @property
    def bucket(self) -> str:
        """Return S3 bucket name."""
        return self._bucket

    @property
    def endpoint(self) -> str:
        """Return S3 bucket endpoint."""
        return self._endpoint

    def on_create(self, dataset: "Dataset") -> None:
        """Hook to perform provider-specific actions on a newly-created dataset."""
        credentials = self.get_credentials()
        prompt_for_credentials(credentials)
        storage = self.get_storage(credentials=credentials)

        # NOTE: The underlying rclone tool cannot tell if a directory within a S3 bucket exists or not
        if not storage.exists(self.uri):
            raise errors.ParameterError(f"S3 bucket '{self.bucket}' doesn't exists.")

        project_context.repository.add_ignored_pattern(pattern=str(dataset.get_datadir()))


class S3Credentials(ProviderCredentials):
    """S3-specific credentials."""

    def __init__(self, provider: S3Provider):
        super().__init__(provider=provider)

    @staticmethod
    def get_credentials_names() -> Tuple[str, ...]:
        """Return a tuple of the required credentials for a provider."""
        return "Access Key ID", "Secret Access Key"

    @property
    def provider(self) -> S3Provider:
        """Return the associated provider instance."""
        return cast(S3Provider, self._provider)

    def get_credentials_section_name(self) -> str:
        """Get section name for storing credentials.

        NOTE: This methods should be overridden by subclasses to allow multiple credentials per providers if needed.
        """
        return f"{self.provider.bucket}.{self.provider.endpoint.lower()}"


def parse_s3_uri(uri: str) -> Tuple[str, str, str]:
    """Extract endpoint, bucket name, and path within the bucket from a given URI.

    NOTE: We only support s3://<hostname>/<bucket-name>/<path> at the moment.
    """
    parsed_uri = urllib.parse.urlparse(uri)

    hostname = parsed_uri.netloc
    path = parsed_uri.path.strip("/")
    bucket, _, path = path.partition("/")

    if parsed_uri.scheme.lower() != "s3":
        raise errors.ParameterError(
            f"Invalid S3 scheme: {uri}.\nValid format is 's3://<hostname>/<bucket-name>/<path>'", show_prefix=False
        )
    if not hostname:
        raise errors.ParameterError(
            f"Hostname is missing in S3 URI: {uri}.\nValid format is 's3://<hostname>/<bucket-name>/<path>'",
            show_prefix=False,
        )
    if not bucket:
        raise errors.ParameterError(
            f"Bucket name is missing in S3 URI: {uri}.\nValid format is 's3://<hostname>/<bucket-name>/<path>'",
            show_prefix=False,
        )

    return hostname, bucket, path.strip("/")



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Graph view model."""

import collections
import html
import io
import json
from enum import Enum
from typing import Dict, Iterator, List, Optional, cast

import pyld
import rdflib
from rdflib import ConjunctiveGraph, Graph
from rdflib.query import ResultRow
from rdflib.tools.rdf2dot import LABEL_PROPERTIES, NODECOLOR, rdf2dot


class DotFormat(Enum):
    """Enum of valid Dot output formats."""

    FULL = 1
    FULL_LANDSCAPE = 2
    DEBUG = 3


class GraphViewModel:
    """ViewModel for a JSON-LD knowledge graph.

    Args:
        graph(List[Dict]): The flattened JSON-LD data as a list of Python dictionaries.
    """

    def __init__(self, graph: List[dict]) -> None:
        self._graph = graph

    def as_jsonld(self) -> List[dict]:
        """Get the JSON-LD representation as a list of dictionaries.

        Returns:
            List[Dict]: The Python list of dictionary representation of the graph.
        """
        return self._graph

    def as_jsonld_string(self, indentation: Optional[int] = 2) -> str:
        """Get the JSON-LD representation as a string.

        Args:
            indentation(int, optional): The indentation to use for pretty-printing,
                set to `None` for no pretty printing (Default value = 2).
        Returns:
            str: The flattened JSON-LD string representation of the graph.
        """

        output = pyld.jsonld.flatten(self._graph)
        return json.dumps(output, indent=indentation)

    def as_rdflib_graph(self) -> Graph:
        """Get the graph as an RDFLib graph.

        Returns:
            Graph: The RDFLib graph.
        """
        data = json.dumps(pyld.jsonld.expand(self._graph))
        graph = ConjunctiveGraph().parse(data=data, format="json-ld")
        return graph

    def as_nt_string(self) -> str:
        """Get the graph as a string in nt format.

        Returns:
            str: The nt string representation of the graph.
        """
        return self.as_rdflib_graph().serialize(format="nt")

    def as_rdf_string(self) -> str:
        """Get the graph as a string in rdf+xml format.

        Returns:
            str: The rdf string representation of the graph.
        """
        return self.as_rdflib_graph().serialize(format="application/rdf+xml")

    def as_dot_string(self, format: DotFormat = DotFormat.FULL) -> str:
        """Get the graph as a Graphviz Dot string.

        Args:
            format(DotFormat): The format to use (Default value = `DotFormat.SIMPLE`).
        Returns:
            str: The Dot string representation of the graph.
        """
        graph = self.as_rdflib_graph()

        graph.bind("prov", "http://www.w3.org/ns/prov#")
        graph.bind("foaf", "http://xmlns.com/foaf/0.1/")
        graph.bind("schema", "http://schema.org/")
        graph.bind("renku", "https://swissdatasciencecenter.github.io/renku-ontology/")

        output = io.StringIO()

        try:
            if format == DotFormat.DEBUG:
                rdf2dot(graph, output)
                return output.getvalue()

            output.write('digraph { \n node [ fontname="DejaVu Sans" ] ; \n ')

            if format == DotFormat.FULL_LANDSCAPE:
                output.write('rankdir="LR" \n')

            self._rdfdot_full(graph, output)

            return output.getvalue()
        finally:
            output.close()

    def _rdfdot_full(self, graph: Graph, output: io.TextIOBase) -> None:
        """Write a full Dot graph representation to output.

        Args:
            graph(Graph): RDFLib graph.
            output(io.TextIOBase): The output stream to write to.
        """
        types = collections.defaultdict(set)
        fields = collections.defaultdict(set)
        nodes: Dict[str, str] = {}

        def node(x):
            """Return a name of the given node."""
            return nodes.setdefault(x, f"node{len(nodes)}")

        def label(x, g):
            """Generate a label for the node."""
            for labelProp in LABEL_PROPERTIES:
                label_ = g.value(x, labelProp)
                if label_:
                    return label_

            try:
                return g.namespace_manager.compute_qname(x)[2]
            except Exception:
                return x

        def formatliteral(literal, g):
            """Format and escape literal."""
            v = html.escape(literal)
            if literal.datatype:
                return f"&quot;{v}&quot;^^{qname(literal.datatype, g)}"
            elif literal.language:
                return f"&quot;{v}&quot;@{literal.language}"
            return "&quot;%s&quot;" % v

        def qname(x, g):
            """Compute qname."""
            try:
                q = g.compute_qname(x)
                return q[0] + ":" + q[2]
            except Exception:
                return x

        def color(p):
            """Choose node color."""
            return "BLACK"

        # filter out nodes and edges created for directories
        sparql = """
        SELECT ?s ?p ?o
        WHERE {
            ?s ?p ?o
            MINUS {
                ?s rdf:type prov:Collection.
            }
            MINUS {
                VALUES ?exclude { prov:wasInformedBy prov:influenced rdf:label }
                ?s ?exclude ?o.
            }
        }
        """

        for s, p, o in cast(Iterator[ResultRow], graph.query(sparql)):
            sn = node(s)
            if p == rdflib.RDFS.label:
                continue
            # inject the type predicate into the node itself
            if p == rdflib.RDF.type:
                types[sn].add((qname(p, graph), html.escape(o)))
                continue
            # add the project membership to the node
            if p == rdflib.term.URIRef("schema:isPartOf"):
                fields[sn].add((qname(p, graph), html.escape(o)))
                continue

            if isinstance(o, (rdflib.URIRef, rdflib.BNode)):
                on = node(o)
                opstr = '\t%s -> %s [ color=%s, label=< <font point-size="12" ' 'color="#336633">%s</font> > ] ;\n'
                output.write(opstr % (sn, on, color(p), qname(p, graph)))
            else:
                fields[sn].add((qname(p, graph), formatliteral(o, graph)))

        for u, n in nodes.items():
            output.write(f"# {u} {n}\n")
            f = [
                '<tr><td align="left"><b>%s</b></td><td align="left">' "<b>%s</b></td></tr>" % x
                for x in sorted(types[n])
            ]
            f += ['<tr><td align="left">%s</td><td align="left">%s</td></tr>' % x for x in sorted(fields[n])]
            opstr = (
                '%s [ shape=none, color=%s label=< <table color="#666666"'
                ' cellborder="0" cellspacing="0" border="1"><tr>'
                '<td colspan="2" bgcolor="grey"><B>%s</B></td></tr><tr>'
                '<td href="%s" bgcolor="#eeeeee" colspan="2">'
                '<font point-size="12" color="#6666ff">%s</font></td>'
                "</tr>%s</table> > ] \n"
            )
            output.write(opstr % (n, NODECOLOR, label(u, graph), u, u, "".join(f)))

        output.write("}\n")



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service template create project controller."""
import os
import shutil
from typing import Any, Dict, Optional, cast

from marshmallow import EXCLUDE

from renku.command.init import create_from_template_local_command
from renku.core import errors
from renku.core.image import ImageObjectRequest
from renku.core.template.template import fetch_templates_source
from renku.core.util.contexts import renku_project_context
from renku.domain_model.template import Template
from renku.infrastructure.repository import Repository
from renku.ui.service.config import CACHE_UPLOADS_PATH, MESSAGE_PREFIX
from renku.ui.service.controllers.api.abstract import ServiceCtrl
from renku.ui.service.controllers.api.mixins import RenkuOperationMixin
from renku.ui.service.controllers.utils.datasets import set_url_for_uploaded_images
from renku.ui.service.errors import UserProjectCreationError, UserTemplateInvalidError
from renku.ui.service.serializers.templates import ProjectTemplateRequest, ProjectTemplateResponseRPC
from renku.ui.service.utils import new_repo_push
from renku.ui.service.views import result_response
from renku.version import __version__, is_release


class TemplatesCreateProjectCtrl(ServiceCtrl, RenkuOperationMixin):
    """Template create project controller."""

    REQUEST_SERIALIZER = ProjectTemplateRequest()
    RESPONSE_SERIALIZER = ProjectTemplateResponseRPC()

    def __init__(self, cache, user_data, request_data):
        """Construct a templates read manifest controller."""
        self.ctx = cast(
            Dict[str, Any],
            TemplatesCreateProjectCtrl.REQUEST_SERIALIZER.load({**user_data, **request_data}, unknown=EXCLUDE),
        )
        self.ctx["commit_message"] = f"{MESSAGE_PREFIX} init {self.ctx['project_name']}"
        super().__init__(cache, user_data, request_data)

        self.template: Optional[Template] = None

    @property
    def context(self):
        """Controller context."""
        return self.ctx

    @property
    def default_metadata(self):
        """Default metadata for project creation."""

        metadata = {
            "__template_source__": self.ctx["template_git_url"],
            "__template_ref__": self.ctx["ref"],
            "__template_id__": self.ctx["identifier"],
            "__namespace__": self.ctx["project_namespace"],
            "__repository__": self.ctx["project_repository"],
            "__sanitized_project_name__": self.ctx["project_name_stripped"],
            "__project_slug__": self.ctx["project_slug"],
            "__project_description__": self.ctx["project_description"],
        }

        cli_version = os.environ.get("RENKU_PROJECT_DEFAULT_CLI_VERSION") or __version__
        if is_release(cli_version):
            metadata["__renku_version__"] = cli_version

        return metadata

    @property
    def git_user(self) -> Dict[str, str]:
        """Extract git user from the user data."""
        return {
            "email": self.user_data["email"],
            "name": self.user_data["fullname"],
        }

    def setup_new_project(self):
        """Setup new project for initialization."""
        # TODO: Request attribute naming on create project and read manifest is not consistent.
        new_project_data = {
            "clone_depth": self.ctx["depth"],
            "git_url": self.ctx["new_project_url"],
            "name": self.ctx["project_name"],
            "slug": self.ctx["project_name_stripped"],
            "description": self.ctx["project_description"],
            "fullname": self.ctx["fullname"],
            "email": self.ctx["email"],
            "owner": self.ctx["project_namespace"],
            "token": self.ctx["token"],
            "initialized": True,
            "image": self.ctx["image"],
        }
        project = self.cache.make_project(self.user, new_project_data)

        new_project_path = project.abs_path
        if new_project_path.exists():
            shutil.rmtree(new_project_path)

        new_project_path.mkdir(parents=True, exist_ok=True)

        return project

    def setup_template(self):
        """Reads template manifest."""
        templates_source = fetch_templates_source(source=self.ctx["template_git_url"], reference=self.ctx["ref"])
        identifier = self.ctx["identifier"]
        try:
            self.template = templates_source.get_template(id=identifier, reference=None)
        except errors.TemplateNotFoundError as e:
            raise UserProjectCreationError(
                error_message=f"the template '{identifier}' does not exist in the target template's repository"
            ) from e
        except errors.InvalidTemplateError as e:
            raise UserTemplateInvalidError(error_message=f"the template '{identifier}' is invalid") from e

        repository = Repository(templates_source.path)
        self.template_version = repository.head.commit.hexsha

        # Verify missing parameters
        template_parameters = {p.name: p for p in self.template.parameters}
        provided_parameters = {p["key"]: p["value"] for p in self.ctx["parameters"]}
        missing_keys = list(template_parameters.keys() - provided_parameters.keys())
        missing_keys = [k for k in missing_keys if not template_parameters[k].has_default]
        if len(missing_keys) > 0:
            raise UserProjectCreationError(error_message=f"the template requires a value for '${missing_keys[0]}'")

        return provided_parameters

    def new_project_push(self, project_path):
        """Push new project to the remote."""
        return new_repo_push(project_path, self.ctx["new_project_url_with_auth"])

    def new_project(self):
        """Create new project from template."""
        provided_parameters = self.setup_template()
        assert self.template is not None
        new_project = self.setup_new_project()
        new_project_path = new_project.abs_path

        image = self.ctx.get("image")
        if image:
            user_cache_dir = CACHE_UPLOADS_PATH / self.user.user_id
            set_url_for_uploaded_images(images=[image], cache=self.cache, user=self.user)

            image = ImageObjectRequest(
                content_url=image["content_url"],
                position=0,
                mirror_locally=image.get("mirror_locally", False),
                safe_image_paths=[user_cache_dir],
            )

        with renku_project_context(new_project_path):
            create_from_template_local_command().build().execute(
                self.template.path,
                name=self.ctx["project_name"],
                namespace=self.ctx["project_namespace"],
                metadata=provided_parameters,
                default_metadata=self.default_metadata,
                custom_metadata=self.ctx["project_custom_metadata"],
                template_version=self.template_version,
                immutable_template_files=self.template.immutable_files,
                automated_template_update=self.template.allow_update,
                user=self.git_user,
                initial_branch=self.ctx["initial_branch"],
                commit_message=self.ctx["commit_message"],
                description=self.ctx["project_description"],
                data_dir=self.ctx.get("data_directory"),
                ssh_supported=self.template.ssh_supported,
                image_request=image,
            )

        self.new_project_push(new_project_path)

        new_project.initialized = True
        new_project.save()

        return {
            "url": self.ctx["new_project_url"],
            "namespace": self.ctx["project_namespace"],
            "name": self.ctx["project_name"],
            "slug": self.ctx["project_name_stripped"],
            "project_id": new_project.project_id,
        }

    def renku_op(self):
        """Renku operation for the controller."""
        # NOTE: We leave it empty since it does not execute renku operation.
        pass

    def to_response(self):
        """Execute controller flow and serialize to service response."""
        return result_response(TemplatesCreateProjectCtrl.RESPONSE_SERIALIZER, self.new_project())



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Install and uninstall Git hooks.

Description
~~~~~~~~~~~

The commit hooks are enabled by default to prevent situation when
some output file is manually modified. It also takes care of adding
relevant files to Git LFS and warns of files added to a dataset's
data directory that haven't been added to dataset metadata.


Commands and options
~~~~~~~~~~~~~~~~~~~~

.. rst-class:: cli-reference-commands

.. click:: renku.ui.cli.githooks:githooks
   :prog: renku githooks
   :nested: full


Examples
~~~~~~~~

.. code-block:: console

    $ renku init
    $ renku run echo hello > greeting.txt
    $ edit greeting.txt
    $ git commit greeting.txt
    You are trying to update some output files.

    Modified outputs:
      greeting.txt

    If you are sure, use "git commit --no-verify".

"""

import os
from pathlib import Path

import click

import renku.ui.cli.utils.color as color


@click.group()
def githooks():
    """Manage Git hooks for Renku repository."""


@githooks.command()
@click.option("--force", is_flag=True, help="Override existing hooks.")
def install(force):
    """Install Git hooks."""
    from renku.command.githooks import install_githooks_command
    from renku.ui.cli.utils.callback import ClickCallback

    communicator = ClickCallback()
    result = install_githooks_command().with_communicator(communicator).build().execute(force, path=Path(os.getcwd()))

    warning_messages = result.output
    if warning_messages:
        for message in warning_messages:
            communicator.warn(message)

    click.secho("OK", fg=color.GREEN)


@githooks.command()
def uninstall():
    """Uninstall Git hooks."""
    from renku.command.githooks import uninstall_githooks_command

    uninstall_githooks_command().build().execute()
    click.secho("OK", fg=color.GREEN)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service datasets view."""
from flask import request

from renku.ui.service.config import SERVICE_PREFIX
from renku.ui.service.controllers.datasets_add_file import DatasetsAddFileCtrl
from renku.ui.service.controllers.datasets_create import DatasetsCreateCtrl
from renku.ui.service.controllers.datasets_edit import DatasetsEditCtrl
from renku.ui.service.controllers.datasets_files_list import DatasetsFilesListCtrl
from renku.ui.service.controllers.datasets_import import DatasetsImportCtrl
from renku.ui.service.controllers.datasets_list import DatasetsListCtrl
from renku.ui.service.controllers.datasets_remove import DatasetsRemoveCtrl
from renku.ui.service.controllers.datasets_unlink import DatasetsUnlinkCtrl
from renku.ui.service.views.api_versions import VERSIONS_FROM_V2_2, VersionedBlueprint
from renku.ui.service.views.decorators import accepts_json, optional_identity, requires_cache, requires_identity
from renku.ui.service.views.error_handlers import (
    handle_common_except,
    handle_datasets_unlink_errors,
    handle_datasets_write_errors,
)
from renku.ui.service.views.v1.datasets import add_v1_specific_dataset_endpoints

DATASET_BLUEPRINT_TAG = "datasets"
dataset_blueprint = VersionedBlueprint(DATASET_BLUEPRINT_TAG, __name__, url_prefix=SERVICE_PREFIX)


@dataset_blueprint.route(
    "/datasets.list", methods=["GET"], provide_automatic_options=False, versions=VERSIONS_FROM_V2_2
)
@handle_common_except
@requires_cache
@optional_identity
def list_datasets_view(user_data, cache):
    """
    List all datasets in a project.

    ---
    get:
      description: List all datasets in a project.
      parameters:
        - in: query
          schema: DatasetListRequest
      responses:
        200:
          description: Listing of all datasets in a project.
          content:
            application/json:
              schema: DatasetListResponseRPC
      tags:
        - datasets
    """
    return DatasetsListCtrl(cache, user_data, dict(request.args)).to_response()


@dataset_blueprint.route(
    "/datasets.files_list", methods=["GET"], provide_automatic_options=False, versions=VERSIONS_FROM_V2_2
)
@handle_common_except
@requires_cache
@optional_identity
def list_dataset_files_view(user_data, cache):
    """
    List files in a dataset.

    ---
    get:
      description: List files in a dataset.
      parameters:
        - in: query
          schema: DatasetFilesListRequest
      responses:
        200:
          description: Listing of all files in a dataset.
          content:
            application/json:
              schema: DatasetFilesListResponseRPC
      tags:
        - datasets
    """
    return DatasetsFilesListCtrl(cache, user_data, dict(request.args)).to_response()


@dataset_blueprint.route(
    "/datasets.add", methods=["POST"], provide_automatic_options=False, versions=VERSIONS_FROM_V2_2
)
@handle_common_except
@handle_datasets_write_errors
@accepts_json
@requires_cache
@requires_identity
def add_file_to_dataset_view(user_data, cache):
    """
    Add the uploaded file to a cloned repository.

    ---
    post:
      description: Add the uploaded file to a cloned repository.
      requestBody:
        content:
          application/json:
            schema: DatasetAddRequest
      responses:
        200:
          description: Details of the added files.
          content:
            application/json:
              schema: DatasetAddResponseRPC
      tags:
        - datasets
    """
    return DatasetsAddFileCtrl(cache, user_data, dict(request.json)).to_response()  # type: ignore


@dataset_blueprint.route(
    "/datasets.create", methods=["POST"], provide_automatic_options=False, versions=VERSIONS_FROM_V2_2
)
@handle_common_except
@handle_datasets_write_errors
@accepts_json
@requires_cache
@requires_identity
def create_dataset_view(user_data, cache):
    """
    Create a new dataset in a project.

    ---
    post:
      description: Create a new dataset in a project.
      requestBody:
        content:
          application/json:
            schema: DatasetCreateRequest
      responses:
        200:
          description: Properties of the created dataset.
          content:
            application/json:
              schema: DatasetCreateResponseRPC
      tags:
        - datasets
    """
    return DatasetsCreateCtrl(cache, user_data, dict(request.json)).to_response()  # type: ignore


@dataset_blueprint.route(
    "/datasets.remove", methods=["POST"], provide_automatic_options=False, versions=VERSIONS_FROM_V2_2
)
@handle_common_except
@accepts_json
@requires_cache
@requires_identity
def remove_dataset_view(user_data, cache):
    """
    Remove a dataset from a project.

    ---
    post:
      description: Remove a dataset from a project.
      requestBody:
        content:
          application/json:
            schema: DatasetRemoveRequest
      responses:
        200:
          description: Details of the removed dataset.
          content:
            application/json:
              schema: DatasetRemoveResponseRPC
      tags:
        - datasets
    """
    return DatasetsRemoveCtrl(cache, user_data, dict(request.json)).to_response()  # type: ignore


@dataset_blueprint.route(
    "/datasets.import", methods=["POST"], provide_automatic_options=False, versions=VERSIONS_FROM_V2_2
)
@handle_common_except
@accepts_json
@requires_cache
@requires_identity
def import_dataset_view(user_data, cache):
    """
    Import a dataset view.

    ---
    post:
      description: Import a dataset into a project.
      requestBody:
        content:
          application/json:
            schema: DatasetImportRequest
      responses:
        200:
          description: Details of the dispatched import dataset job.
          content:
            application/json:
              schema: DatasetImportResponseRPC
      tags:
        - datasets
    """
    return DatasetsImportCtrl(cache, user_data, dict(request.json)).to_response()  # type: ignore


@dataset_blueprint.route(
    "/datasets.edit", methods=["POST"], provide_automatic_options=False, versions=VERSIONS_FROM_V2_2
)
@handle_common_except
@handle_datasets_write_errors
@accepts_json
@requires_cache
@requires_identity
def edit_dataset_view(user_data, cache):
    """
    Edit dataset metadata view.

    Not passing a field leaves it unchanged.

    ---
    post:
      description: Edit dataset metadata.
      requestBody:
        content:
          application/json:
            schema: DatasetEditRequest
      responses:
        200:
          description: Status of the requested dataset edits.
          content:
            application/json:
              schema: DatasetEditResponseRPC
      tags:
        - datasets
    """
    return DatasetsEditCtrl(cache, user_data, dict(request.json)).to_response()  # type: ignore


@dataset_blueprint.route(
    "/datasets.unlink", methods=["POST"], provide_automatic_options=False, versions=VERSIONS_FROM_V2_2
)
@handle_common_except
@handle_datasets_unlink_errors
@accepts_json
@requires_cache
@requires_identity
def unlink_file_view(user_data, cache):
    """
    Unlink a file from a dataset view.

    ---
    post:
      description: Unlink a file from a dataset.
      requestBody:
        content:
          application/json:
            schema: DatasetUnlinkRequest
      responses:
        200:
          description: Details of the unlinked files.
          content:
            application/json:
              schema: DatasetUnlinkResponseRPC
      tags:
        - datasets
    """
    return DatasetsUnlinkCtrl(cache, user_data, dict(request.json)).to_response()  # type: ignore


dataset_blueprint = add_v1_specific_dataset_endpoints(dataset_blueprint)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service cache views."""
from flask import jsonify, request

from renku.ui.service.config import SERVICE_PREFIX
from renku.ui.service.controllers.cache_files_delete_chunks import DeleteFileChunksCtrl
from renku.ui.service.controllers.cache_files_upload import UploadFilesCtrl
from renku.ui.service.controllers.cache_list_uploaded import ListUploadedFilesCtrl
from renku.ui.service.controllers.cache_migrate_project import MigrateProjectCtrl
from renku.ui.service.controllers.cache_migrations_check import MigrationsCheckCtrl
from renku.ui.service.gateways.gitlab_api_provider import GitlabAPIProvider
from renku.ui.service.gateways.repository_cache import LocalRepositoryCache
from renku.ui.service.jobs.cleanup import cache_files_cleanup
from renku.ui.service.views.api_versions import (
    ALL_VERSIONS,
    VERSIONS_FROM_V1_1,
    VERSIONS_FROM_V2_0,
    VERSIONS_FROM_V2_1,
    VersionedBlueprint,
)
from renku.ui.service.views.decorators import accepts_json, optional_identity, requires_cache, requires_identity
from renku.ui.service.views.error_handlers import (
    handle_common_except,
    handle_migration_read_errors,
    handle_migration_write_errors,
)
from renku.ui.service.views.v1.cache import add_v1_specific_cache_endpoints

CACHE_BLUEPRINT_TAG = "cache"
cache_blueprint = VersionedBlueprint("cache", __name__, url_prefix=SERVICE_PREFIX)


@cache_blueprint.route("/cache.files_list", methods=["GET"], provide_automatic_options=False, versions=ALL_VERSIONS)
@handle_common_except
@requires_cache
@requires_identity
def list_uploaded_files_view(user_data, cache):
    """
    List uploaded files.

    ---
    get:
      description: List uploaded files ready to be added to projects.
      responses:
        200:
          description: "Return a list of files."
          content:
            application/json:
              schema: FileListResponseRPC
      tags:
        - cache
    """
    return ListUploadedFilesCtrl(cache, user_data).to_response()


@cache_blueprint.route("/cache.files_upload", methods=["POST"], provide_automatic_options=False, versions=ALL_VERSIONS)
@handle_common_except
@requires_cache
@requires_identity
def upload_file_view(user_data, cache):
    """
    Upload a file or archive of files.

    ---
    post:
      description: Upload a file or archive of files.
      parameters:
        - in: query
          schema: FileUploadRequest
      requestBody:
        content:
          multipart/form-data:
            schema:
              type: object
              properties:
                file:
                  type: string
                  format: binary
      responses:
        200:
          description: List of uploaded files.
          content:
            application/json:
              schema: FileUploadResponseRPC
      tags:
        - cache
    """
    return UploadFilesCtrl(cache, user_data, request).to_response()


@cache_blueprint.route(
    "/cache.files_delete_chunks", methods=["POST"], provide_automatic_options=False, versions=ALL_VERSIONS
)
@handle_common_except
@accepts_json
@requires_cache
@requires_identity
def delete_file_chunks_view(user_data, cache):
    """
    Delete chunks from a chunked upload.

    ---
    post:
      description: Delete chunks from a chunked upload.
      parameters:
        - in: query
          schema: FileChunksDeleteRequest
      responses:
        200:
          description: Status of deletion request.
          content:
            application/json:
              schema: FileChunksDeleteResponseRPC
      tags:
        - cache
    """
    return DeleteFileChunksCtrl(cache, user_data, dict(request.json)).to_response()  # type: ignore


@cache_blueprint.route("/cache.migrate", methods=["POST"], provide_automatic_options=False, versions=VERSIONS_FROM_V1_1)
@handle_common_except
@handle_migration_write_errors
@accepts_json
@requires_cache
@requires_identity
def migrate_project_view(user_data, cache):
    """
    Migrate a project.

    ---
    post:
      description: Migrate a project.
      requestBody:
        content:
          application/json:
            schema: ProjectMigrateRequest
      responses:
        200:
          description: Migration status.
          content:
            application/json:
              schema: ProjectMigrateResponseRPC
      tags:
        - cache
    """
    return MigrateProjectCtrl(cache, user_data, dict(request.json)).to_response()  # type: ignore


@cache_blueprint.route(
    "/cache.migrations_check", methods=["GET"], provide_automatic_options=False, versions=VERSIONS_FROM_V2_0
)
@handle_common_except
@handle_migration_read_errors
@requires_cache
@optional_identity
def migration_check_project_view(user_data, cache):
    """
    Retrieve migration information for a project.

    ---
    get:
      description: Retrieve migration information for a project.
      parameters:
        - in: query
          schema: ProjectMigrationCheckRequest
      responses:
        200:
          description: Information about required migrations for the project.
          content:
            application/json:
              schema: ProjectMigrationCheckResponseRPC
      tags:
        - cache
    """
    return MigrationsCheckCtrl(cache, user_data, dict(request.args), GitlabAPIProvider()).to_response()


@cache_blueprint.route("/cache.cleanup", methods=["GET"], provide_automatic_options=False, versions=VERSIONS_FROM_V2_1)
@handle_common_except
@handle_migration_read_errors
@requires_cache
@optional_identity
def cache_cleanup(user_data, cache):
    """
    Cleanup local project cache.

    ---
    get:
      description: Retrieve migration information for a project.
      responses:
        200:
          description: Information about required migrations for the project.
          content:
            application/json:
              schema: CacheCleanupResponseRPC
      tags:
        - cache
    """
    LocalRepositoryCache().evict_expired()
    cache_files_cleanup()
    return jsonify({"result": "ok"})


cache_blueprint = add_v1_specific_cache_endpoints(cache_blueprint)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service templates controller."""

from renku.ui.service.controllers.templates_create_project import TemplatesCreateProjectCtrl
from renku.ui.service.controllers.templates_read_manifest import TemplatesReadManifestCtrl
from renku.ui.service.serializers.v1.templates import ManifestTemplatesRequest_v2_2, ProjectTemplateRequest_v2_2


class TemplatesCreateProjectCtrl_v2_2(TemplatesCreateProjectCtrl):
    """V2.2 create project controller."""

    REQUEST_SERIALIZER = ProjectTemplateRequest_v2_2()  # type: ignore


class TemplatesReadManifestCtrl_v2_2(TemplatesReadManifestCtrl):
    """V2.2 read manifest controller."""

    REQUEST_SERIALIZER = ManifestTemplatesRequest_v2_2()  # type: ignore



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Move or rename a file, a directory, or a symlink."""

import os
from pathlib import Path
from typing import List, Optional

from pydantic import validate_arguments

from renku.command.command_builder import inject
from renku.command.command_builder.command import Command
from renku.core import errors
from renku.core.dataset.dataset import move_files
from renku.core.dataset.datasets_provenance import DatasetsProvenance
from renku.core.interface.dataset_gateway import IDatasetGateway
from renku.core.storage import track_paths_in_storage, untrack_paths_from_storage
from renku.core.util import communication
from renku.core.util.metadata import is_protected_path
from renku.core.util.os import get_relative_path, is_subpath
from renku.domain_model.project_context import project_context


def move_command():
    """Command to move or rename a file, a directory, or a symlink."""
    return Command().command(_move).require_migration().require_clean().with_database(write=True).with_commit()


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def _move(sources: List[str], destination: str, force: bool, verbose: bool, to_dataset: Optional[str]):
    """Move files and check repository for potential problems.

    Args:
        sources(List[str]): Source file(s) to move.
        destination(str): Destination to move files to.
        force(bool): Whether or not to overwrite destination files.
        verbose(bool): Toggle verbose output.
        to_dataset(Optional[str]): Target dataset to move files into.
    """
    repository = project_context.repository

    absolute_destination = _get_absolute_path(destination)
    absolute_sources = [_get_absolute_path(src) for src in sources]

    if to_dataset:
        target_dataset = DatasetsProvenance().get_by_slug(to_dataset, strict=True)
        if not is_subpath(absolute_destination, _get_absolute_path(target_dataset.get_datadir())):
            raise errors.ParameterError(
                f"Destination {destination} must be in {target_dataset.get_datadir()} when moving to a dataset."
            )

    is_rename = len(absolute_sources) == 1 and (
        not absolute_destination.exists() or (absolute_destination.is_file() and absolute_sources[0].is_file())
    )

    files = {
        path: _get_dst(path, src, absolute_destination, is_rename)
        for src in absolute_sources
        for path in _traverse_path(src)
        if not path.is_dir()
    }

    if not files:
        raise errors.ParameterError("There are no files to move.")
    if not force:
        _check_existing_destinations(files.values())
        _warn_about_dataset_files(files)

    # NOTE: we don't check if source and destination are the same or if multiple sources are moved to the same
    # destination; git mv will check those and we raise if git mv fails.

    _warn_about_ignored_destinations(files.values())

    if not absolute_destination.exists() and not absolute_destination.is_symlink():
        if is_rename:
            absolute_destination.parent.mkdir(parents=True, exist_ok=True)
        else:
            absolute_destination.mkdir(parents=True, exist_ok=True)

    try:
        repository.move(*sources, destination=destination, force=force)
    except errors.GitCommandError as e:
        raise errors.OperationError(f"Git operation failed: {e}")

    # Handle symlinks
    for src, dst in files.items():
        if dst.is_symlink():
            target = src.parent / os.readlink(dst)
            dst.unlink()
            Path(dst).symlink_to(os.path.relpath(target, start=os.path.dirname(dst)))

    files_to_untrack = (str(src.relative_to(project_context.path)) for src in files)
    untrack_paths_from_storage(*files_to_untrack)
    # NOTE: Warn about filter after untracking from LFS to avoid warning about LFS filters
    _warn_about_git_filters(files)
    track_paths_in_storage(*[dst for dst in files.values() if not dst.is_dir()])

    # NOTE: Force-add to include possible ignored files
    repository.add(*files.values(), force=True)

    move_files(files=files, to_dataset_slug=to_dataset)

    if verbose:
        _show_moved_files(project_context.path, files)


def _traverse_path(path):
    """Recursively yield all files and directories within a path.

    Args:
        path: Root path.

    Returns:
        Iterator of all paths in root path.
    """
    path = Path(path)

    if path.is_dir():
        yield from path.rglob("*")
    else:
        yield path


def _get_dst(path, src_root, dst_root, is_rename):
    parent = "." if is_rename else src_root.name
    return dst_root / parent / os.path.relpath(path, start=src_root)


def _get_absolute_path(path):
    """Resolve path and raise if path is outside the repo or is protected.

    Args:
        path: Path to make absolute.

    Returns:
        Absolute path.
    """
    abs_path = Path(os.path.abspath(path))

    if is_protected_path(abs_path):
        raise errors.ParameterError(f"Path '{path}' is protected.")

    try:
        abs_path.relative_to(project_context.path)
    except ValueError:
        raise errors.ParameterError(f"Path '{path}' is outside the project.")

    return abs_path


def _check_existing_destinations(destinations):
    existing = set()
    for dst in destinations:
        if dst.exists() or dst.is_symlink():
            existing.add(str(dst.relative_to(os.getcwd())))

    if not existing:
        return

    existing_str = "\n\t".join(existing)
    raise errors.ParameterError(
        f"The following move target exist, use '--force' flag to overwrite them:\n\t{existing_str}"
    )


def _warn_about_ignored_destinations(destinations):
    ignored = project_context.repository.get_ignored_paths(*destinations)
    if ignored:
        ignored_str = "\n\t".join(str(Path(p).relative_to(project_context.path)) for p in ignored)
        communication.warn(f"The following moved path match .gitignore:\n\t{ignored_str}")


def _warn_about_git_filters(files):
    """Check if there are any git attributes for files including LFS.

    Args:
        files: Files to check.
    """
    repository = project_context.repository

    src_attrs = []
    dst_attrs = []

    for path, attrs in repository.get_attributes(*files).items():
        src = Path(path)
        dst = files[src].relative_to(project_context.path)
        src = src.relative_to(project_context.path)
        attrs_text = ""
        for name, value in attrs.items():
            if value == "unset":
                attrs_text += f" -{name}"
            elif value == "set":
                attrs_text += f" {name}"
            else:
                attrs_text += f" {name}={value}"

        src_attrs.append(f"{str(src)}{attrs_text}")
        dst_attrs.append(f"{str(dst)}{attrs_text}")

    if src_attrs:
        src_attrs_str = "\n\t".join(src_attrs)
        dst_attrs_str = "\n\t".join(dst_attrs)
        communication.warn(
            f"There are custom git attributes for the following files:\n\t{src_attrs_str}\n"
            f"You need to edit '.gitattributes' and add the following:\n\t{dst_attrs_str}"
        )


@inject.autoparams()
def _warn_about_dataset_files(files, dataset_gateway: IDatasetGateway):
    """Check if any of the files are part of a dataset.

    Args:
        files: Files to check.
        dataset_gateway(IDatasetGateway): Injected dataset gateway.
    """
    found = []
    for dataset in dataset_gateway.get_all_active_datasets():
        for src, dst in files.items():
            relative_src = get_relative_path(src, project_context.path)
            if not relative_src:
                continue

            found_file = dataset.find_file(relative_src)
            if not found_file:
                continue
            if not found_file.is_external and not is_subpath(dst, project_context.path / dataset.get_datadir()):
                found.append(str(src))

    if not found:
        return

    found_str = "\n\t".join(found)
    communication.confirm(
        msg="You are trying to move dataset files out of a datasets data directory. "
        f"These files will be removed from the source dataset:\n\t{found_str}",
        abort=True,
        warning=True,
    )


def _show_moved_files(project_path, files):
    for path in sorted(files):
        src = path.relative_to(project_path)
        dst = files[path].relative_to(project_path)
        communication.echo(f"{src} -> {dst}")



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""External storage interface."""

import abc
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Optional, Union

if TYPE_CHECKING:
    from renku.core.dataset.providers.api import CloudStorageProviderType, ProviderCredentials


@dataclass
class FileHash:
    """The hash for a file at a specific location."""

    uri: str
    path: str
    size: Optional[int]  # Size in bytes
    hash: Optional[str]


class IStorageFactory(abc.ABC):
    """Interface to get a cloud storage."""

    @staticmethod
    @abc.abstractmethod
    def get_storage(
        storage_scheme: str,
        provider: "CloudStorageProviderType",
        credentials: "ProviderCredentials",
        configuration: Dict[str, str],
    ) -> "IStorage":
        """Return a storage that handles provider.

        Args:
            storage_scheme(str): Storage name.
            provider(CloudStorageProviderType): The backend provider.
            credentials(ProviderCredentials): Credentials for the provider.
            configuration(Dict[str, str]): Storage-specific configuration that are passed to the IStorage implementation

        Returns:
            An instance of IStorage.
        """
        raise NotImplementedError


class IStorage(abc.ABC):
    """Interface for the external storage handler."""

    def __init__(
        self,
        storage_scheme: str,
        provider: "CloudStorageProviderType",
        credentials: "ProviderCredentials",
        provider_configuration: Dict[str, str],
    ):
        self._storage_scheme: str = storage_scheme
        self._provider: "CloudStorageProviderType" = provider
        self._credentials: "ProviderCredentials" = credentials
        self._provider_configuration: Dict[str, str] = provider_configuration

    @property
    def credentials(self) -> "ProviderCredentials":
        """Return the provider credentials for this storage handler."""
        return self._credentials

    @property
    def provider(self) -> "CloudStorageProviderType":
        """Return the dataset provider for this storage handler."""
        return self._provider

    @property
    def storage_scheme(self) -> str:
        """Storage's URI scheme."""
        return self._storage_scheme

    @abc.abstractmethod
    def download(self, uri: str, destination: Union[Path, str]) -> None:
        """Download data from ``uri`` to ``destination``."""
        raise NotImplementedError

    @abc.abstractmethod
    def exists(self, uri: str) -> bool:
        """Checks if a remote storage URI exists."""
        raise NotImplementedError

    @abc.abstractmethod
    def get_configurations(self) -> Dict[str, str]:
        """Get required configurations to access the storage."""
        raise NotImplementedError

    @abc.abstractmethod
    def get_hashes(self, uri: str, hash_type: str = "md5") -> List[FileHash]:
        """Get the hashes of all files at the uri."""
        raise NotImplementedError

    @abc.abstractmethod
    def is_directory(self, uri: str) -> bool:
        """Return True if URI points to a directory."""
        raise NotImplementedError

    @abc.abstractmethod
    def mount(self, path: Union[Path, str]) -> None:
        """Mount the provider's URI to the given path."""
        raise NotImplementedError

    @abc.abstractmethod
    def upload(self, source: Union[Path, str], uri: str) -> None:
        """Upload data from ``source`` to ``uri``."""
        raise NotImplementedError



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service project lock status controller."""

import portalocker

from renku.core import errors
from renku.ui.service.cache.models.project import Project
from renku.ui.service.controllers.api.abstract import ServiceCtrl
from renku.ui.service.controllers.api.mixins import RenkuOperationMixin
from renku.ui.service.errors import IntermittentProjectIdError
from renku.ui.service.serializers.project import ProjectLockStatusRequest, ProjectLockStatusResponseRPC
from renku.ui.service.utils import normalize_git_url
from renku.ui.service.views import result_response


class ProjectLockStatusCtrl(ServiceCtrl, RenkuOperationMixin):
    """Controller for project lock status endpoint."""

    REQUEST_SERIALIZER = ProjectLockStatusRequest()
    RESPONSE_SERIALIZER = ProjectLockStatusResponseRPC()

    def __init__(self, cache, user_data, request_data):
        """Construct a project edit controller."""
        self.ctx = self.REQUEST_SERIALIZER.load(request_data)

        super().__init__(cache, user_data, request_data)

        if "git_url" in self.ctx:
            self.ctx["git_url"] = normalize_git_url(self.ctx["git_url"])

    @property
    def context(self):
        """Controller operation context."""
        return self.ctx

    def get_lock_status(self) -> bool:
        """Return True if a project is write-locked."""
        if "project_id" in self.context:
            try:
                project = self.cache.get_project(self.user, self.context["project_id"])
            except IntermittentProjectIdError:
                return False
        elif "git_url" in self.context and "user_id" in self.user_data:
            try:
                project = Project.get(
                    (Project.user_id == self.user_data["user_id"]) & (Project.git_url == self.context["git_url"])
                )
            except ValueError:
                return False
        else:
            raise errors.RenkuException("context does not contain `project_id` or `git_url` or missing `user_id`")

        try:
            with project.read_lock(timeout=self.ctx["timeout"]):
                return False
        except (portalocker.LockException, portalocker.AlreadyLocked):
            return True

    def renku_op(self):
        """Renku operation for the controller."""
        # NOTE: We leave it empty since it does not execute renku operation.
        pass

    def to_response(self):
        """Execute controller flow and serialize to service response."""
        is_locked = self.get_lock_status()
        return result_response(self.RESPONSE_SERIALIZER, data={"locked": is_locked})



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Image model."""

from pathlib import Path
from typing import Union
from urllib.parse import urlparse

from renku.core.util.urls import get_path, is_remote
from renku.infrastructure.immutable import Slots


class ImageObject(Slots):
    """Represents a schema.org ``ImageObject``."""

    __slots__ = ("content_url", "id", "position")

    id: str
    content_url: str
    position: int

    def __init__(self, *, content_url: Union[str, Path], id: str, position: int):
        id = get_path(id)
        super().__init__(content_url=str(content_url), position=position, id=id)

    @staticmethod
    def generate_id(owner_id: str, position: int) -> str:
        """Generate @id field."""
        return f"{owner_id}/images/{position}"

    @property
    def is_absolute(self):
        """Whether content_url is an absolute or relative url."""
        return bool(urlparse(self.content_url).netloc)

    @property
    def is_remote(self) -> bool:
        """Return True if the URI isn't on the local filesystem."""
        return is_remote(self.content_url)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""DOI API integration."""

import urllib
from pathlib import Path

from renku.core import errors
from renku.core.dataset.providers.api import ImporterApi, ImportProviderInterface, ProviderApi, ProviderPriority
from renku.core.util.doi import extract_doi, is_doi

DOI_BASE_URL = "https://dx.doi.org"


class DOIProvider(ProviderApi, ImportProviderInterface):
    """`doi.org <http://doi.org>`_ registry API provider."""

    priority = ProviderPriority.HIGHER
    name = "DOI"
    is_remote = True

    def __init__(self, uri: str, headers=None, timeout=3):
        super().__init__(uri=uri)

        self.timeout = timeout
        self.headers = headers if headers is not None else {"accept": "application/vnd.citationstyles.csl+json"}

    @staticmethod
    def supports(uri) -> bool:
        """Whether or not this provider supports a given URI."""
        return bool(is_doi(uri))

    def get_importer(self, **kwargs) -> "DOIImporter":
        """Get import manager."""
        from renku.core.util import requests

        def query(doi):
            """Retrieve metadata for given doi."""
            doi = extract_doi(doi)
            url = make_doi_url(doi)

            response = requests.get(url, headers=self.headers)

            if response.status_code != 200:
                raise LookupError(f"record not found. Status: {response.status_code}")

            return response

        def serialize(response):
            """Serialize HTTP response for DOI."""
            json_data = response.json()
            data = {key.replace("-", "_").lower(): value for key, value in json_data.items()}
            try:
                return DOIImporter(**data)
            except TypeError:
                raise errors.DatasetImportError("doi metadata could not be serialized")

        query_response = query(self.uri)
        return serialize(query_response)


class DOIImporter(ImporterApi):
    """Response from `doi.org <http://doi.org>`_ for DOI metadata."""

    def __init__(
        self,
        id,
        doi,
        url,
        abstract=None,
        author=None,
        categories=None,
        container_title=None,
        contributor=None,
        copyright=None,
        issued=None,
        language=None,
        publisher=None,
        title=None,
        type=None,
        version=None,
    ):
        super().__init__(uri=url, original_uri=url)

        self.id = id
        self.doi = doi

        self.abstract = abstract
        self.author = author
        self.categories = categories
        self.container_title = container_title
        self.contributor = contributor
        self.copyright = copyright
        self.issued = issued
        self.language = language
        self.publisher = publisher
        self.name = title
        self.type = type
        self._version = version

    @property
    def version(self) -> str:
        """Get record version."""
        return self._version

    @property
    def latest_uri(self) -> str:
        """Get URI of the latest version."""
        return self.uri

    def fetch_provider_dataset(self):
        """Deserialize this record to a ``ProviderDataset``."""
        raise NotImplementedError

    def is_latest_version(self) -> bool:
        """Check if record is at last possible version."""
        return True

    def download_files(self, destination: Path, extract: bool):
        """Download dataset files from the remote provider."""
        raise NotImplementedError

    def tag_dataset(self, name: str) -> None:
        """Create a tag for the dataset ``name`` if the remote dataset has a tag/version."""
        raise NotImplementedError

    def copy_extra_metadata(self, new_dataset) -> None:
        """Copy provider specific metadata once the dataset is created."""
        raise NotImplementedError


def make_doi_url(doi):
    """Create URL to access DOI metadata."""
    parsed_url = urllib.parse.urlparse(doi)
    if parsed_url.scheme == "doi":
        parsed_url = parsed_url._replace(scheme="")
        doi = parsed_url.geturl()
    return urllib.parse.urljoin(DOI_BASE_URL, doi)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Track provenance of data created by executing programs.

Description
~~~~~~~~~~~

Track the execution of your command line scripts. This will enable detection of:

* arguments (flags),
* string and integer options,
* input files or directories if linked to existing paths in the repository,
* output files or directories if modified or created while running the command.

It will create a ``Plan`` (Workflow Template) that can be reused and a ``Run``
which is a record of a past workflow execution for provenance purposes. Refer
to the :ref:`cli-workflow` documentation for more details on this distinction.

Commands and options
~~~~~~~~~~~~~~~~~~~~

.. rst-class:: cli-reference-commands

.. click:: renku.ui.cli.run:run
   :prog: renku run
   :nested: full

Examples
~~~~~~~~

.. code-block:: console

    $ renku run --name <plan name> -- <console command>

.. note:: If there were uncommitted changes in the repository, then the
   ``renku run`` command fails. See :program:`git status` for details.

.. warning:: If executed command/script has similar arguments to ``renku run``
    (e.g. ``--input``) they will be treated as ``renku run`` arguments. To
    avoid this, put a ``--`` separator between ``renku run`` and the
    command/script.

.. warning:: Input and output paths can only be detected if they are passed as
   arguments to ``renku run``.

.. warning:: Circular dependencies are not supported for ``renku run``. See
   :ref:`circular-dependencies` for more details.

.. warning:: When using output redirection in ``renku run`` on Windows (with
   `` > file`` or `` 2> file``), all Renku errors and messages are redirected
   as well and ``renku run`` produces no output on the terminal. On Linux,
   this is detected by renku and only the output of the command to be run is
   actually redirected. Renku specific messages such as errors get printed to
   the terminal as usual and don't get redirected.

.. cheatsheet::
   :group: Running
   :command: $ renku run --name <name> <command> [--input <in_file>...] [--output <out_file>...]
   :description: Execute a <command> with Renku tracking inputs and outputs. Input and output files
                 are automatically detected from the command string. Creates a workflow template
                 named <name>. With --input and/or --output: Manually specify input or output files to track.
   :target: rp

Detecting input paths
~~~~~~~~~~~~~~~~~~~~~

Any path passed as an argument to ``renku run``, which was not changed during
the execution, is identified as an input path. The identification only works if
the path associated with the argument matches an existing file or directory
in the repository.

The detection might not work as expected if:

* a file is **modified** during the execution. In this case it will be stored
  as an **output**;
* a path is not passed as an argument to ``renku run``.

.. topic:: Specifying auxiliary inputs (``--input``)

   You can specify extra inputs to your program explicitly by using the
   ``--input`` option. This is useful for specifying hidden dependencies
   that don't appear on the command line. Explicit inputs must exist before
   execution of ``renku run`` command. This option is not a replacement for
   the arguments that are passed on the command line. Files or directories
   specified with this option will not be passed as input arguments to the
   script.
   You can specify ``--input name=path`` or just ``--input path``, the former
   of which would also set the name of the input on the resulting Plan.

   For example, ``renku run --input infile=data.csv -- python script.py data.csv outfile``
   would force Renku to detect ``data.csv`` as an input file and set the name
   of the input to ``infile``.
   Similarly, ``renku run --input infile=data.csv -- python script.py``
   would let Renku know that ``script.py`` reads the file ``data.csv`` even
   though it does not show up on the command line.

.. topic:: Specifying auxiliary parameters (``--param``)

   You can specify extra parameters to your program explicitly by using the
   ``--param`` option. This is useful for getting Renku to consider a
   parameter as just a string even if it matches a file name in the project.
   This option is not a replacement for the arguments that are passed on the
   command line.
   You can specify ``--param name=value`` or just ``--param value``, the former
   of which would also set the name of the parameter on the resulting Plan.

   For example, ``renku run --param my-param=hello -- python script.py hello outfile``
   would force Renku to detect ``hello`` as the value of a string parameter
   with name ``my-param`` even if there is a file called ``hello`` present on the
   filesystem.

.. topic:: Disabling input detection (``--no-input-detection``)

    Input paths detection can be disabled by passing ``--no-input-detection``
    flag to ``renku run``. In this case, only the directories/files that are
    passed as explicit input are considered to be file inputs. Those passed via
    command arguments are ignored unless they are in the explicit inputs list.
    This only affects files and directories; command options and flags are
    still treated as inputs.

.. topic:: Disabling parameter detection (``--no-parameter-detection``)

    Inputs that aren't files or directories are automatically detected as
    parameters in ``renku run``. You can disable this feature by passing the
    ``--no-parameter-detection`` flag, which completely ignores them on the
    workflow. You can still manually specify parameters using ``--param``
    arguments mentioned above or using the ``renku.api.Parameter`` class in
    Python code.

.. note:: ``renku run`` prints the generated plan after execution if you pass
    ``--verbose`` to it. You can check the generated plan to verify that the
    execution was done as you intended. The plan will always be printed to
    ``stderr`` even if it's directed to a file.

Detecting output paths
~~~~~~~~~~~~~~~~~~~~~~

Any path **modified** or **created** during the execution will be added as an
output.

Because the output path detection is based on the Git repository state after
the execution of ``renku run`` command, it is good to have a basic
understanding of the underlying principles and limitations of tracking
files in Git.

Git tracks not only the paths in a repository, but also the content stored in
those paths. Therefore:

* a recreated file with the same content is not considered an output file,
  but instead is kept as an input;
* file moves are detected based on their content and can cause problems;
* directories cannot be empty.

.. note:: When in doubt whether the outputs will be detected, remove all
  outputs using ``git rm <path>`` followed by ``git commit`` before running
  the ``renku run`` command.

.. topic:: Command does not produce any files (``--no-output``)

   If the program does not produce any outputs, the execution ends with an
   error:

   .. code-block:: text

      Error: There are not any detected outputs in the repository.

   You can specify the ``--no-output`` option to force tracking of such
   an execution.

.. cheatsheet::
   :group: Running
   :command: $ renku run --name <name> <command> --no-output
   :description: Run a <command> that produces no output.
   :target: rp

.. topic:: Specifying outputs explicitly (``--output``)

   You can specify expected outputs of your program explicitly by using the
   ``--output`` option. These output must exist after the execution of the
   ``renku run`` command. However, they do not need to be modified by
   the command.
   You can specify ``--output name=path`` or just `--output path``, the former
   of which would also set the name of the output on the resulting Plan.

   For instance, ``renku run --output result=result.txt -- python script.py -o result.txt``
   would force Renku to treat the file ``result.txt`` as an output of the
   workflow and set the name of the output to ``result``.
   Similarly, ``renku run --output result=result.txt -- python script.py``
   would let Renku know about ``result.txt`` created by ``script.py`` even
   though it does not show up on the command line command. Though Renku should
   automatically detect these cases under normal circumstances.

.. topic:: Disabling output detection (``--no-output-detection``)

    Output paths detection can be disabled by passing ``--no-output-detection``
    flag to ``renku run``. When disabled, only the directories/files that are
    passed as explicit output are considered to be outputs and those passed via
    command arguments are ignored.

.. cli-run-std

Detecting standard streams
~~~~~~~~~~~~~~~~~~~~~~~~~~

Often the program expect inputs as a standard input stream. This is detected
and recorded in the tool specification when invoked by ``renku run cat < A``.

Similarly, both redirects to standard output and standard error output can be
done when invoking a command:

.. code-block:: console

   $ renku run grep "test" B > C 2> D

.. warning:: Detecting inputs and outputs from pipes ``|`` is not supported.

Specifying inputs and outputs programmatically
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Sometimes the list of inputs and outputs are not known before execution of the
program. For example, a program might accept a date range as input and access
all files within that range during its execution.

To address this issue, the program can dump a mapping of input and output files
that it is accessing in ``inputs.yml`` and ``outputs.yml``. This YAML file
should be of the format

.. code-block:: YAML

   name1: path1
   name2: path2

where name is the user-defined name of the input/output and path is the path.
When the program is finished, Renku will look for existence of these two files
and adds their content to the list of explicit inputs and outputs. Renku will
then delete these two files.

By default, Renku looks for these two files in ``.renku/tmp`` directory. One
can change this default location by setting ``RENKU_INDIRECT_PATH``
environment variable. When set, it points to a sub-directory within the
``.renku/tmp`` directory where ``inputs.yml`` and ``outputs.yml`` reside.

Exit codes
~~~~~~~~~~

All Unix commands return a number between 0 and 255 which is called an
"exit code". In case other numbers are returned, they are treated modulo 256
(-10 is equivalent to 246, 257 is equivalent to 1). The exit-code 0 represents
a *success* and non-zero exit-code indicates a *failure*.

Therefore the command specified after ``renku run`` is expected to return
exit-code 0. If the command returns different exit code, you can specify them
with ``--success-code=<INT>`` parameter.

.. code-block:: console

   $ renku run --success-code=1 --no-output fail

.. _circular-dependencies:

Circular Dependencies
~~~~~~~~~~~~~~~~~~~~~

Circular dependencies are not supported in ``renku run``. This means you cannot
use the same file or directory as both an input and an output in the same step,
for instance reading from a file as input and then appending to it is not
allowed. Since renku records all steps of an analysis workflow in a dependency
graph and it allows you to update outputs when an input changes, this would
lead to problems with circular dependencies. An update command would change the
input again, leading to renku seeing it as a changed input, which would run
update again, and so on, without ever stopping.

Due to this, the renku dependency graph has to be *acyclic*. So instead of
appending to an input file or writing an output file to the same directory
that was used as an input directory, create new files or write to other
directories, respectively.

.. _workflow-definition-file:

Workflow Definition File
~~~~~~~~~~~~~~~~~~~~~~~~

Instead of using ``renku run`` to track your workflows, you can pass a workflow
definition file to renku for execution and tracking. A workflow definition file
or workflow file contains definition of each individual command as execution
*steps*. A step's definition includes the command that will be executed along
with lists of all its inputs, outputs, and parameters that are used in the
command. The following shows a workflow file with one step:

.. code-block:: console

    name: workflow-file
    steps:
      head:
        command: head -n 10 data/collection/models.csv data/collection/colors.csv > intermediate
        inputs:
          - models:
              path: data/collection/models.csv
          - colors:
              path: data/collection/colors.csv
        outputs:
          temporary-result:
            path: intermediate
        parameters:
          n:
            prefix: -n
            value: 10

The step *head* in this workflow file, has two inputs, one output, and one
parameter. All these arguments are given a name for better understanding of
their purpose. The same workflow file can be simplified to the following
format:

.. code-block:: console

    name: workflow-file
    steps:
      head:
        command: head -n 10 data/collection/models.csv data/collection/colors.csv > intermediate
        inputs:
          - data/collection/models.csv
          - data/collection/colors.csv
        outputs:
          - intermediate
        parameters:
          - -n
          - 10

Although the latter format is more concise it's recommended to use the former
format since it's more readable and has a more clear definition. You can
provide a description for each of the elements in the workflow file. You can
also have a set of keywords for each step and for the workflow file. The
following listing shows a more complete definition of the same workflow file:

.. code-block:: console

    name: workflow-file
    description: A sample workflow file used for testing
    keywords:
      - workflow file
      - v1
    steps:
      head:
        command: head -n 10 data/collection/models.csv data/collection/colors.csv > intermediate
        description: first stage of the pipeline
        success_codes:
          - 0
          - 127
        keywords:
          - preprocessing
          - first step
        inputs:
          - models:
              description: all available model numbers
              path: data/collection/models.csv
          - colors:
              path: data/collection/colors.csv
        outputs:
          temporary-result:
            description: temporary intermediate result that won't be saved
            path: intermediate
        parameters:
          n:
            description: number of lines to print
            prefix: -n
            value: 10

.. topic:: Running a workflow file

    The ``renku run`` command can be used to execute a workflow file:

    .. code-block:: console

       $ renku run workflow-file.yml

    .. note:: It's recommended to use a ``.yml`` or ``.yaml`` extension for your
        workflow files. Renku uses this as a hint to guess that you want to run a
        workflow file (and not an executable script). You can force execution of
        a workflow file by passing a ``--file`` flag to the ``renku run`` command.

    Upon running, Renku executes each step of the workflow file and creates
    corresponding plans for them. The effect is exactly the same as running each
    step using ``renku run`` in the command line.

    .. note:: The order of execution of steps is determined by the dependencies
        between the steps and not by the order that they are defined in the
        workflow file.

    Automatic detection of inputs and outputs is disabled when running a workflow
    file. The reason is that the list of inputs and outputs are already given in
    the workflow file and there is no need for automatic detection. This also
    allows running workflow files in a dirty repository which is not possible when
    running individual commands. Moreover, once the execution is over, only the
    outputs that have been generated by the workflow file will be committed.

    You can also select a subset of steps for execution by passing their names
    after the workflow file. The following command runs only two steps named
    *step-2* and *step-7* in an example workflow files:

    .. code-block:: console

       $ renku run workflow-file.yml step-2 step-7

.. topic:: Argument templates in workflow files

    In the workflow definition file, each command's input, output, or parameter
    is defined in two places: One in the command and one in the list of arguments.
    The reason for this duplication is that Renku needs to know the type of each
    argument (whether it's an input, output, or parameter) as well as the order
    that each argument has in the command. This redundant definition of arguments
    is inconvenient for large workflow files and is error-prone.

    To address this issue, Renku supports argument templates for the workflow
    files. To use templates, you need to assign a name to each argument and then
    use this name to refer to each argument in the step's command. The following
    listing shows the first workflow file rewritten using templates:

    .. code-block:: console

        name: workflow-file
        steps:
          head:
            command: head $n $models $colors > $temporary-result
            inputs:
              - models:
                  path: data/collection/models.csv
              - colors:
                  path: data/collection/colors.csv
            outputs:
              temporary-result:
                path: intermediate
            parameters:
              n:
                prefix: -n
                value: 10

    As is shown, you can use the name of each argument (i.e. input, output, or
    parameter) prefixed with a ``$`` to refer to an argument.

    In addition, you can use ``$inputs``, ``$outputs``, or ``$parameters`` to refer
    to the list of all inputs, outputs, or parameters respectively:

    .. code-block:: console

        ...
            command: cmd $parameters $inputs $outputs
        ...

    .. note:: If input or output paths or parameter values contain a ``$`` you
        must replace them with ``$$`` in the command. This doesn't need to be
        done in the list of arguments.

.. topic:: ``--dry-run`` and ``--no-commit`` flags

    By passing the ``--dry-run`` flag to the ``renku run`` command, you can
    instruct Renku to only print the order of execution of the steps without
    actually running any of them. Note that this option is valid only for running
    workflow files and it is ignored when running a single command.

    The ``--no-commit`` flags causes Renku to run the workflow file but it won't
    create a commit after the execution. Renku also won't create any metadata in
    this case.

    .. code-block:: console

       $ renku run --no-commit workflow-file.yml
"""

import os
from pathlib import Path

import click
from lazy_object_proxy import Proxy

from renku.command.run import run_workflow_file_command
from renku.core import errors
from renku.core.plugin.workflow_file_parser import read_workflow_file
from renku.core.util.os import is_subpath
from renku.core.workflow.workflow_file import get_workflow_file_inputs_and_outputs
from renku.domain_model.project_context import project_context
from renku.ui.cli.utils.callback import ClickCallback
from renku.ui.cli.utils.plugins import available_workflow_providers
from renku.ui.cli.utils.terminal import print_workflow_file


@click.command(context_settings=dict(ignore_unknown_options=True))
@click.option("--name", help="A name for the workflow step.")
@click.option("--description", help="Workflow step's description.")
@click.option("--keyword", multiple=True, help="List of keywords for the workflow.")
@click.option("explicit_inputs", "--input", multiple=True, help="Force a path to be considered as an input.")
@click.option("explicit_outputs", "--output", multiple=True, help="Force a path to be considered an output.")
@click.option("explicit_parameters", "--param", multiple=True, help="Force a string to be considered a parameter.")
@click.option("--no-output", is_flag=True, default=False, help="Allow command without output files.")
@click.option("--no-input-detection", is_flag=True, default=False, help="Disable auto-detection of inputs.")
@click.option("--no-output-detection", is_flag=True, default=False, help="Disable auto-detection of outputs.")
@click.option("--no-parameter-detection", is_flag=True, default=False, help="Disable auto-detection of parameters.")
@click.option(
    "--success-code",
    "success_codes",
    type=int,
    multiple=True,
    callback=lambda _, __, values: [int(value) % 256 for value in values],
    help="Allowed command exit-code.",
)
@click.option("--isolation", is_flag=True, default=False, help="Invoke the given command in isolation.")
@click.option("--file", is_flag=True, default=False, help="Force running of a workflow file.")
@click.option("--verbose", is_flag=True, default=False, help="Print generated plan after the execution.")
@click.option(
    "--creator",
    "creators",
    default=None,
    multiple=True,
    type=click.UNPROCESSED,
    help="Creator's name, email, and affiliation. Accepted format is 'Forename Surname <email> [affiliation]'.",
)
@click.option("--dry-run", is_flag=True, help="Show what would have been executed in a workflow file")
@click.option("--no-commit", is_flag=True, help="Don't update metadata after the execution and don't create a commit.")
@click.option(
    "--provider",
    default=None,
    show_default=False,
    type=click.Choice(Proxy(available_workflow_providers), case_sensitive=False),
    help="The workflow engine to use for executing workflow files.",
)
@click.argument("command_line", nargs=-1, metavar="<COMMAND> or <WORKFLOW FILE>", required=True, type=click.UNPROCESSED)
def run(
    name,
    description,
    keyword,
    explicit_inputs,
    explicit_outputs,
    explicit_parameters,
    no_output,
    no_input_detection,
    no_output_detection,
    no_parameter_detection,
    success_codes,
    isolation,
    file,
    command_line,
    verbose,
    creators,
    dry_run,
    no_commit,
    provider,
):
    """Tracking work on a specific problem."""
    from renku.command.run import run_command_line_command
    from renku.core.util.metadata import construct_creators
    from renku.ui.cli.utils.terminal import print_plan

    communicator = ClickCallback()

    def is_workflow_file() -> bool:
        """Some heuristics to guess if the first argument is a workflow file or not."""
        if file:
            return True
        if not command_line:
            return False

        path = Path(command_line[0])
        if path.suffix.lower() in [".yml", ".yaml"] and not os.access(path, os.X_OK):
            return True
        if not path.is_file() or not is_subpath(path=path, base=project_context.path):
            return False

        try:
            content = read_workflow_file(path=path)
        except (errors.ParseError, errors.ParameterError):
            return False
        else:
            # NOTE: A single string is also a valid YAML, so, we check that the file contains a dict
            return isinstance(content, dict)

    if is_workflow_file():
        # NOTE: Check other flags and warn if they are set since they don't have any effect
        if (
            name
            or description
            or keyword
            or explicit_inputs
            or explicit_outputs
            or explicit_parameters
            or no_output
            or no_input_detection
            or no_output_detection
            or no_parameter_detection
            or success_codes
            or isolation
            or creators
        ):
            communicator.warn("All flags other than '--file', '--verbose', '--dry-run', and 'no-commit' are ignored")

        path = command_line[0]
        steps = command_line[1:]
        no_commit = no_commit or dry_run

        # NOTE: Read the workflow file to get list of generated files that should be committed
        if no_commit:
            workflow_file = None
            commit_only = None
        else:
            workflow_file = read_workflow_file(path=path, parser="renku")
            commit_only = (
                [path]
                + get_workflow_file_inputs_and_outputs(workflow_file=workflow_file, steps=steps)
                + [str(project_context.metadata_path)]
            )

        provider = provider or "local"

        result = (
            run_workflow_file_command(no_commit=no_commit, commit_only=commit_only)
            .with_communicator(communicator)
            .build()
            .execute(path=path, steps=steps, dry_run=dry_run, workflow_file=workflow_file, provider=provider)
        )

        if dry_run:
            for line in result.output[1]:
                communicator.echo(line)
        elif verbose:
            print_workflow_file(result.output[0])
    else:
        # NOTE: ``dry-run`` and ``no-commit`` should be passed only with WFF
        if dry_run:
            communicator.warn("'--dry-run' flag is valid only when running workflow files")
        if no_commit:
            communicator.warn("'--no-commit' flag is valid only when running workflow files")
        if provider:
            communicator.warn("'--provider' flag is valid only when running workflow files")

        command = run_command_line_command()

        if isolation:
            command = command.with_git_isolation()

        if creators:
            creators, _ = construct_creators(creators)

        result = (
            command.with_communicator(communicator)
            .build()
            .execute(
                name=name,
                description=description,
                keyword=keyword,
                explicit_inputs=explicit_inputs,
                explicit_outputs=explicit_outputs,
                explicit_parameters=explicit_parameters,
                no_output=no_output,
                no_input_detection=no_input_detection,
                no_output_detection=no_output_detection,
                no_parameter_detection=no_parameter_detection,
                success_codes=success_codes,
                command_line=command_line,
                creators=creators,
            )
        )

        if verbose:
            plan = result.output
            print_plan(plan, err=True)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Template management."""

import json
import os
import re
import shutil
import tempfile
from enum import Enum, IntEnum, auto
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

from packaging.version import Version

from renku.core import errors
from renku.core.util import communication
from renku.core.util.git import clone_repository
from renku.core.util.util import to_semantic_version, to_string
from renku.domain_model.project_context import project_context
from renku.domain_model.template import (
    TEMPLATE_MANIFEST,
    RenderedTemplate,
    Template,
    TemplateMetadata,
    TemplateParameter,
    TemplatesManifest,
    TemplatesSource,
    hash_template_file,
    update_dockerfile_content,
)
from renku.infrastructure.repository import Repository

try:
    import importlib_resources  # type:ignore
except ImportError:
    import importlib.resources as importlib_resources  # type:ignore

from renku.domain_model.project import Project, ProjectTemplateMetadata

TEMPLATE_KEEP_FILES = ["readme.md", "readme.rst", "readme.txt", "readme"]
TEMPLATE_INIT_APPEND_FILES = [".gitignore"]


class TemplateAction(Enum):
    """Types of template rendering."""

    INITIALIZE = auto()
    SET = auto()
    UPDATE = auto()


class FileAction(IntEnum):
    """Types of operation when copying a template to a project."""

    APPEND = 1
    CREATE = 2
    DELETED = 3
    IGNORE_IDENTICAL = 4
    IGNORE_UNCHANGED_REMOTE = 5
    KEEP = 6
    OVERWRITE = 7
    RECREATE = 8
    UPDATE_DOCKERFILE = 9


def fetch_templates_source(source: Optional[str], reference: Optional[str]) -> TemplatesSource:
    """Fetch a template."""
    if reference and not source:
        raise errors.ParameterError("Can't use a template reference without specifying a template source")

    return (
        EmbeddedTemplates.fetch(source, reference)
        if is_renku_template(source)
        else RepositoryTemplates.fetch(source, reference)
    )


def is_renku_template(source: Optional[str]) -> bool:
    """Return if template comes from Renku."""
    return not source or source.lower() == "renku"


def write_template_checksum(checksums: Dict):
    """Write templates checksum file for a project."""
    project_context.template_checksums_path.parent.mkdir(parents=True, exist_ok=True)

    with open(project_context.template_checksums_path, "w") as checksum_file:
        json.dump(checksums, checksum_file)


def read_template_checksum() -> Dict[str, str]:
    """Read templates checksum file for a project."""
    if has_template_checksum():
        with open(project_context.template_checksums_path) as checksum_file:
            return json.load(checksum_file)

    return {}


def has_template_checksum() -> bool:
    """Return if project has a templates checksum file."""
    return os.path.exists(project_context.template_checksums_path)


def copy_template_to_project(
    rendered_template: RenderedTemplate, project: "Project", actions: Dict[str, FileAction], cleanup=True
):
    """Update project files and metadata from a template."""

    def copy_template_metadata_to_project():
        """Update template-related metadata in a project."""
        write_template_checksum(rendered_template.checksums)

        project.template_metadata = ProjectTemplateMetadata(
            template_id=rendered_template.template.id,
            template_source=rendered_template.template.source,
            template_ref=rendered_template.template.reference,
            template_version=rendered_template.template.version,
            immutable_template_files=rendered_template.template.immutable_files.copy(),
            metadata=json.dumps(rendered_template.metadata),
            ssh_supported=rendered_template.template.ssh_supported,
        )

    actions_mapping: Dict[FileAction, Tuple[str, str]] = {
        FileAction.APPEND: ("append", "Appending to"),
        FileAction.CREATE: ("copy", "Initializing"),
        FileAction.DELETED: ("", "Ignoring deleted file"),
        FileAction.IGNORE_IDENTICAL: ("", "Ignoring identical file"),
        FileAction.IGNORE_UNCHANGED_REMOTE: ("", "Ignoring unchanged template file"),
        FileAction.KEEP: ("", "Keeping"),
        FileAction.OVERWRITE: ("copy", "Overwriting"),
        FileAction.RECREATE: ("copy", "Recreating deleted file"),
        FileAction.UPDATE_DOCKERFILE: ("dockerfile", "Updating"),
    }

    for relative_path, action in get_sorted_actions(actions=actions).items():
        source = rendered_template.path / relative_path
        destination = project_context.path / relative_path

        operation, message = actions_mapping[action]
        communication.echo(f"{message} {relative_path} ...")

        if not operation:
            continue

        try:
            destination.parent.mkdir(parents=True, exist_ok=True)

            if operation == "copy":
                shutil.copy(source, destination, follow_symlinks=False)
            elif operation == "append":
                destination.write_text(destination.read_text() + "\n" + source.read_text())
            elif operation == "dockerfile":
                update_dockerfile_content(source=source, destination=destination)
            else:
                raise errors.TemplateUpdateError(f"Unknown operation '{operation}'")
        except OSError as e:
            # TODO: Use a general cleanup strategy: https://github.com/SwissDataScienceCenter/renku-python/issues/736
            if cleanup:
                repository = project_context.repository
                repository.reset(hard=True)
                repository.clean()

            raise errors.TemplateUpdateError(f"Cannot write to '{destination}'") from e

    copy_template_metadata_to_project()


def get_sorted_actions(actions: Dict[str, FileAction]) -> Dict[str, FileAction]:
    """Return a sorted actions list."""
    return {k: v for k, v in sorted(actions.items(), key=lambda i: (i[1], i[0]))}


def get_file_actions(
    rendered_template: RenderedTemplate, template_action: TemplateAction, interactive
) -> Dict[str, FileAction]:
    """Render a template regarding files in a project."""
    from renku.core.template.usecase import is_dockerfile_updated_by_user

    if interactive and not communication.has_prompt():
        raise errors.ParameterError("Cannot use interactive mode with no prompt")

    old_checksums = read_template_checksum()
    try:
        immutable_files = project_context.project.template_metadata.immutable_template_files or []
    except (AttributeError, ValueError):  # NOTE: Project is not set
        immutable_files = []

    def should_append(path: str):
        return path.lower() in TEMPLATE_INIT_APPEND_FILES

    def should_keep(path: str):
        return path.lower() in TEMPLATE_KEEP_FILES

    def get_action_for_initialize(relative_path: str, destination: Path) -> FileAction:
        if not destination.exists():
            return FileAction.CREATE
        elif should_append(relative_path):
            return FileAction.APPEND
        elif should_keep(relative_path):
            return FileAction.KEEP
        else:
            return FileAction.OVERWRITE

    def get_action_for_set(relative_path: str, destination: Path, new_checksum: Optional[str]) -> FileAction:
        """Decide what to do with a template file."""
        current_checksum = hash_template_file(relative_path=relative_path, absolute_path=destination)

        if not destination.exists():
            return FileAction.CREATE
        if new_checksum == current_checksum:
            return FileAction.IGNORE_IDENTICAL
        elif interactive:
            overwrite = communication.confirm(f"Overwrite {relative_path}?", default=True)
            return FileAction.OVERWRITE if overwrite else FileAction.KEEP
        elif should_keep(relative_path):
            return FileAction.KEEP
        else:
            return FileAction.OVERWRITE

    def get_action_for_update(
        relative_path: str, destination: Path, old_checksum: Optional[str], new_checksum: Optional[str]
    ) -> FileAction:
        """Decide what to do with a template file."""
        current_checksum = hash_template_file(relative_path=relative_path, absolute_path=destination)
        local_changes = current_checksum != old_checksum
        remote_changes = new_checksum != old_checksum
        file_exists = destination.exists()
        file_deleted = not file_exists and old_checksum is not None

        if not file_deleted and new_checksum == current_checksum:
            return FileAction.IGNORE_IDENTICAL
        if not file_exists and not file_deleted:
            return FileAction.CREATE
        elif interactive:
            if file_deleted:
                recreate = communication.confirm(f"Recreate deleted {relative_path}?", default=True)
                return FileAction.RECREATE if recreate else FileAction.DELETED
            else:
                overwrite = communication.confirm(f"Overwrite {relative_path}?", default=True)
                return FileAction.OVERWRITE if overwrite else FileAction.KEEP
        elif not remote_changes:
            return FileAction.IGNORE_UNCHANGED_REMOTE
        elif relative_path == "Dockerfile":
            if is_dockerfile_updated_by_user():
                raise errors.TemplateUpdateError("Can't update template as Dockerfile was locally changed.")
            else:
                return FileAction.UPDATE_DOCKERFILE
        elif file_deleted or local_changes:
            if relative_path in immutable_files:
                # NOTE: There are local changes in a file that should not be changed by users, and the file was
                # updated in the template as well. So the template can't be updated.
                raise errors.TemplateUpdateError(
                    f"Can't update template as immutable template file '{relative_path}' has local changes."
                )

            # NOTE: Don't overwrite files that are modified by users
            return FileAction.DELETED if file_deleted else FileAction.KEEP
        else:
            return FileAction.OVERWRITE

    actions: Dict[str, FileAction] = {}

    for relative_path in sorted(rendered_template.get_files()):
        destination = project_context.path / relative_path

        if destination.is_dir():
            raise errors.TemplateUpdateError(
                f"Cannot copy a file '{relative_path}' from template to the directory '{relative_path}'"
            )

        new_checksum = rendered_template.checksums[relative_path]

        if template_action == TemplateAction.INITIALIZE:
            action = get_action_for_initialize(relative_path, destination)
        elif template_action == TemplateAction.SET:
            action = get_action_for_set(relative_path, destination, new_checksum=new_checksum)
        else:
            action = get_action_for_update(
                relative_path,
                destination,
                old_checksum=old_checksums.get(relative_path),
                new_checksum=new_checksum,
            )

        actions[relative_path] = action

    return actions


def set_template_parameters(
    template: Template, template_metadata: TemplateMetadata, input_parameters: Dict[str, str], interactive=False
):
    """Set and verify template parameters' values in the template_metadata."""
    if interactive and not communication.has_prompt():
        raise errors.ParameterError("Cannot use interactive mode with no prompt")

    def validate(var: TemplateParameter, val) -> Tuple[bool, Any]:
        try:
            return True, var.convert(val)
        except ValueError as e:
            communication.info(str(e))
            return False, val

    def read_valid_value(var: TemplateParameter, default_value=None):
        """Prompt the user for a template variable and return a valid value."""
        while True:
            variable_type = f", type: {var.type}" if var.type else ""
            enum_values = f", options: {var.possible_values}" if var.possible_values else ""
            default_value = default_value or to_string(var.default)

            val = communication.prompt(
                f"Enter a value for '{var.name}' ({var.description}{variable_type}{enum_values})",
                default=default_value,
                show_default=var.has_default,
            )

            valid, val = validate(var, val)
            if valid:
                return val

    missing_values = []

    for parameter in sorted(template.parameters, key=lambda v: v.name):
        name = parameter.name
        is_valid = True

        if name in input_parameters:  # NOTE: Inputs override other values. No prompt for them in interactive mode
            is_valid, value = validate(parameter, input_parameters[name])
        elif interactive:
            value = read_valid_value(parameter, default_value=template_metadata.metadata.get(name))
        elif name in template_metadata.metadata:
            is_valid, value = validate(parameter, template_metadata.metadata[name])
        elif parameter.has_default:  # Use default value if no value is available in the metadata
            value = parameter.default
        elif communication.has_prompt():
            value = read_valid_value(parameter)
        else:
            missing_values.append(name)
            continue

        if not is_valid:
            if not communication.has_prompt():
                raise errors.TemplateUpdateError(f"Invalid value '{value}' for variable '{name}'")
            template_metadata.metadata[name] = read_valid_value(parameter)
        else:
            template_metadata.metadata[name] = value

    if missing_values:
        missing_values_str = ", ".join(missing_values)
        raise errors.TemplateUpdateError(f"Can't update template, it now requires variable(s): {missing_values_str}")

    # NOTE: Ignore internal variables, i.e. __\w__
    internal_keys = re.compile(r"^__\w+__$")
    metadata_variables = {v for v in template_metadata.metadata if not internal_keys.match(v)} | set(
        input_parameters.keys()
    )
    template_variables = {v.name for v in template.parameters}
    unused_metadata_variables = metadata_variables - template_variables
    if len(unused_metadata_variables) > 0:
        unused_str = "\n\t".join(unused_metadata_variables)
        communication.info(f"These parameters are not used by the template and were ignored:\n\t{unused_str}\n")


class EmbeddedTemplates(TemplatesSource):
    """Represent templates that are bundled with Renku.

    For embedded templates, ``source`` is "renku". In the old versioning scheme, ``version`` is set to the installed
    Renku version and ``reference`` is not set. In the new scheme, both ``version`` and ``reference`` are set to the
    template version.
    """

    @classmethod
    def fetch(cls, source: Optional[str], reference: Optional[str]) -> "EmbeddedTemplates":
        """Fetch embedded Renku templates."""
        from renku import __template_version__

        template_path = importlib_resources.files("renku") / "templates"
        with importlib_resources.as_file(template_path) as templates:
            path = Path(templates)

        return cls(path=path, source="renku", reference=__template_version__, version=__template_version__)

    def get_all_references(self, id) -> List[str]:
        """Return all available references for a template id."""
        template_exists = any(id == t.id or id in t.aliases for t in self.templates)
        return [self.reference] if template_exists and self.reference is not None else []

    def get_latest_reference_and_version(
        self, id: str, reference: Optional[str], version: Optional[str]
    ) -> Optional[Tuple[Optional[str], str]]:
        """Return latest reference and version number of a template."""
        if version is None:
            return None
        elif reference is None or reference != version:  # Old versioning scheme
            return self.reference, self.version

        try:
            current_version = Version(version)
        except ValueError:  # NOTE: version is not a valid SemVer
            return self.reference, self.version
        else:
            return (self.reference, self.version) if current_version < Version(self.version) else (reference, version)

    def get_template(self, id, reference: Optional[str]) -> "Template":
        """Return all available versions for a template id."""
        try:
            return next(t for t in self.templates if id == t.id or id in t.aliases)
        except StopIteration:
            raise errors.TemplateNotFoundError(f"The template with id '{id}' is not available.")


class RepositoryTemplates(TemplatesSource):
    """Represent a local/remote template repository.

    A template repository is checked out at a specific Git reference if one is provided. However, it's still possible to
    get available versions of templates.

    For these templates, ``reference`` is set to whatever user passed as a reference (defaults to remote HEAD if not
    passed) and ``version`` is set to the commit SHA of the reference commit.
    """

    def __init__(self, path, source, reference, version, repository: Repository, skip_validation: bool = False):
        super().__init__(
            path=path, source=source, reference=reference, version=version, skip_validation=skip_validation
        )
        self.repository: Repository = repository

    @classmethod
    def fetch(cls, source: Optional[str], reference: Optional[str]) -> "RepositoryTemplates":
        """Fetch a template repository."""
        ref_str = f"@{reference}" if reference else ""
        communication.echo(f"Fetching template from {source}{ref_str}... ")
        path = Path(tempfile.mkdtemp())

        try:
            repository = clone_repository(url=source, path=path, checkout_revision=reference, install_lfs=False)
        except errors.GitError as e:
            if "Cannot checkout reference" in str(e):
                raise errors.TemplateMissingReferenceError(
                    f"Cannot find the reference '{reference}' in the template repository from {source}"
                ) from e
            raise errors.InvalidTemplateError(f"Cannot clone template repository from {source}") from e

        version = repository.head.commit.hexsha

        return cls(path=path, source=source, reference=reference, version=version, repository=repository)

    def get_all_references(self, id) -> List[str]:
        """Return a list of git tags that are valid SemVer and include a template id."""
        versions = []
        for tag in self.repository.tags:
            tag = str(tag)

            version = to_semantic_version(tag)
            if not version:
                continue

            if self._has_template_at(id, reference=tag):
                versions.append(version)

        return [str(v) for v in sorted(versions)]

    def get_latest_reference_and_version(
        self, id: str, reference: Optional[str], version: Optional[str]
    ) -> Optional[Tuple[Optional[str], str]]:
        """Return latest reference and version number of a template."""
        if version is None:
            return None

        tag = None
        if reference is not None:
            tag = to_semantic_version(reference)

        # NOTE: Assume that a SemVer reference is always a tag
        if tag:
            references = self.get_all_references(id=id)
            return (references[-1], self.version) if len(references) > 0 else None

        # NOTE: Template's reference is a branch or SHA and the latest version is RepositoryTemplates' version
        return reference, self.version

    def _has_template_at(self, id: str, reference: str) -> bool:
        """Return if template id is available at a reference."""
        try:
            content = self.repository.get_content(TEMPLATE_MANIFEST, revision=reference)
            manifest = TemplatesManifest.from_string(content)
        except (errors.FileNotFound, errors.InvalidTemplateError):
            return False
        else:
            return any(id == t.id or id in t.aliases for t in manifest.templates)

    def get_template(self, id, reference: Optional[str]) -> "Template":
        """Return a template at a specific reference."""
        if reference is not None and reference != self.reference:
            try:
                self.repository.checkout(reference=reference)
            except errors.GitError as e:
                raise errors.InvalidTemplateError(f"Cannot find reference '{reference}'") from e
            else:
                self.reference = reference
                self.version = self.repository.head.commit.hexsha

            try:
                manifest = TemplatesManifest.from_path(self.path / TEMPLATE_MANIFEST)
            except errors.InvalidTemplateError as e:
                raise errors.InvalidTemplateError(f"Cannot load template's manifest file at '{reference}'.") from e
            else:
                self.manifest = manifest

        template = next((t for t in self.templates if id == t.id or id in t.aliases), None)
        if template is None:
            reference = reference or "HEAD"
            raise errors.TemplateNotFoundError(f"The template with id '{id}' is not available at '{reference}'.")

        return template



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Running workflows logic."""

import os
import sys
from collections import defaultdict
from io import UnsupportedOperation
from pathlib import Path
from subprocess import call
from typing import Dict, List, NamedTuple, Optional, Set, Union, cast

import click
from pydantic import validate_arguments

from renku.command.command_builder import inject
from renku.command.view_model.plan import PlanViewModel
from renku.core import errors
from renku.core.config import get_value
from renku.core.git import get_mapped_std_streams
from renku.core.interface.activity_gateway import IActivityGateway
from renku.core.interface.plan_gateway import IPlanGateway
from renku.core.storage import check_external_storage, pull_paths_from_storage
from renku.core.util.datetime8601 import local_now
from renku.core.util.git import get_git_user
from renku.core.util.os import get_relative_path_to_cwd, get_relative_paths
from renku.core.util.urls import get_slug
from renku.core.workflow.activity import (
    get_all_modified_and_deleted_activities_and_entities,
    get_downstream_generating_activities,
    is_activity_valid,
)
from renku.core.workflow.plan_factory import PlanFactory
from renku.domain_model.project_context import project_context
from renku.domain_model.provenance.activity import Activity
from renku.domain_model.provenance.agent import Person


class StatusResult(NamedTuple):
    """Represent status of a project.

    A quintuple containing a mapping of stale outputs to modified usages, a mapping of stale activities that have no
    generation to modified usages, a set of modified usages, a set of deleted usages, and a map of modified hidden
    inputs to paths that they generate.
    """

    outdated_outputs: Dict[str, Set[str]]
    outdated_activities: Dict[str, Set[str]]
    modified_inputs: Set[str]
    deleted_inputs: Set[str]
    modified_hidden_inputs: Dict[str, Set[str]]


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def get_status(paths: Optional[List[Union[Path, str]]] = None, ignore_deleted: bool = False) -> StatusResult:
    """Return status of a project.

    Args:
        paths(Optional[List[Union[Path, str]]]): Limit the status to this list of paths (Default value = None).
        ignore_deleted(bool): Whether to ignore deleted generations (Default value = False).

    Returns:
        StatusResult: Status of the project.

    """

    def get_all_generations(activity):
        return (
            get_relative_path_to_cwd(project_context.path / generation.entity.path)
            for generation in activity.generations
        )

    def get_downstream_activities(start_activity, entity, ignore_deleted):
        # NOTE: Add all downstream activities if the modified entity is in paths; otherwise, add only activities that
        # chain-generate at least one of the paths
        generation_paths = [] if not paths or entity.path in paths else paths

        return get_downstream_generating_activities(
            starting_activities={start_activity},
            paths=generation_paths,
            ignore_deleted=ignore_deleted,
            project_path=project_context.path,
        )

    ignore_deleted = ignore_deleted or get_value("renku", "update_ignore_delete")

    modified, deleted, hidden_modified = get_all_modified_and_deleted_activities_and_entities(
        repository=project_context.repository, check_hidden_dependencies=True
    )

    modified = {(a, e) for a, e in modified if is_activity_valid(a)}
    deleted = {(a, e) for a, e in deleted if is_activity_valid(a)}
    hidden_modified = {(a, e) for a, e in hidden_modified if is_activity_valid(a)}

    if not modified and not deleted and not hidden_modified:
        return StatusResult({}, {}, set(), set(), {})

    paths = paths or []
    paths = get_relative_paths(base=project_context.path, paths=[Path.cwd() / p for p in paths])  # type: ignore

    modified_inputs: Set[str] = set()
    stale_outputs: Dict[str, Set[str]] = defaultdict(set)
    stale_activities: Dict[str, Set[str]] = defaultdict(set)
    modified_hidden_inputs: Dict[str, Set[str]] = defaultdict(set)

    for start_activity, entity in modified:
        activities = get_downstream_activities(
            start_activity=start_activity, entity=entity, ignore_deleted=ignore_deleted
        )
        if activities:
            usage_path = get_relative_path_to_cwd(project_context.path / entity.path)
            modified_inputs.add(usage_path)

            for activity in activities:
                if len(activity.generations) == 0:
                    stale_activities[activity.id].add(usage_path)
                else:
                    for generation_path in get_all_generations(activity):
                        stale_outputs[generation_path].add(usage_path)

    deleted_paths = {e.path for _, e in deleted}
    deleted_paths = {
        get_relative_path_to_cwd(project_context.path / d) for d in deleted_paths if not paths or d in paths
    }

    for start_activity, entity in hidden_modified:
        activities = get_downstream_activities(start_activity=start_activity, entity=entity, ignore_deleted=True)
        if activities:
            usage_path = get_relative_path_to_cwd(project_context.path / entity.path)
            for activity in activities:
                for generation_path in get_all_generations(activity):
                    modified_hidden_inputs[usage_path].add(generation_path)

    return StatusResult(
        outdated_outputs=stale_outputs,
        outdated_activities=stale_activities,
        modified_inputs=modified_inputs,
        deleted_inputs=deleted_paths,
        modified_hidden_inputs=modified_hidden_inputs,
    )


def get_valid_plan_name(name: str) -> str:
    """Return a valid plan name based on the passed name."""
    return get_slug(name, invalid_chars=["."], lowercase=False)


def get_valid_parameter_name(name: str) -> str:
    """Return a valid parameter name based on the passed name."""
    return get_valid_plan_name(name)


@inject.autoparams("activity_gateway", "plan_gateway")
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def run_command_line(
    name: Optional[str],
    description: Optional[str],
    keyword: Optional[List[str]],
    explicit_inputs: List[str],
    explicit_outputs: List[str],
    explicit_parameters: List[str],
    no_output: bool,
    no_input_detection: bool,
    no_output_detection: bool,
    no_parameter_detection: bool,
    success_codes: List[int],
    command_line: List[str],
    creators: Optional[List[Person]],
    activity_gateway: IActivityGateway,
    plan_gateway: IPlanGateway,
) -> PlanViewModel:
    """Run command line arguments."""
    # NOTE: validate name as early as possible
    if name:
        valid_name = get_valid_plan_name(name)
        if name != valid_name:
            raise errors.ParameterError(f"Invalid name: '{name}' (Hint: '{valid_name}' is valid).")

        workflows = plan_gateway.get_newest_plans_by_names()
        if name in workflows:
            raise errors.DuplicateWorkflowNameError(f"Duplicate workflow name: Workflow '{name}' already exists.")

    paths = explicit_outputs if no_output_detection else project_context.repository.all_files
    mapped_std = get_mapped_std_streams(paths, streams=("stdout", "stderr"))

    paths = explicit_inputs if no_input_detection else project_context.repository.all_files
    mapped_std_in = get_mapped_std_streams(paths, streams=("stdin",))
    mapped_std.update(mapped_std_in)

    invalid = get_mapped_std_streams(explicit_inputs, streams=("stdout", "stderr"))
    if invalid:
        raise errors.UsageError(
            "Explicit input file cannot be used as stdout/stderr:"
            "\n\t" + click.style("\n\t".join(invalid.values()), fg="yellow") + "\n"
        )

    invalid = get_mapped_std_streams(explicit_outputs, streams=("stdin",))
    if invalid:
        raise errors.UsageError(
            "Explicit output file cannot be used as stdin:"
            "\n\t" + click.style("\n\t".join(invalid.values()), fg="yellow") + "\n"
        )

    system_stdout = None
    system_stderr = None

    # /dev/tty is a virtual device that points to the terminal
    # of the currently executed process
    try:
        with open("/dev/tty", "w"):
            tty_exists = True
    except OSError:
        tty_exists = False

    try:
        stdout_redirected = "stdout" in mapped_std
        stderr_redirected = "stderr" in mapped_std

        if tty_exists:
            # if renku was called with redirected stdout/stderr, undo the
            # redirection here so error messages can be printed normally
            if stdout_redirected:
                system_stdout = open("/dev/tty", "w")
                old_stdout = sys.stdout
                sys.stdout = system_stdout

            if stderr_redirected:
                system_stderr = open("/dev/tty", "w")
                old_stderr = sys.stderr
                sys.stderr = system_stderr

        working_dir = str(project_context.path)

        def parse_explicit_definition(entries, type):
            result = [tuple(e.split("=", maxsplit=1)[::-1]) if "=" in e else (e, None) for e in entries]

            if not result:
                return result

            values, names = zip(*result)

            if len(values) != len(set(values)):
                raise errors.UsageError(f"Cannot specify the same explicit {type} value twice.")

            names = [n for n in names if n]

            if names and len(names) != len(set(names)):
                raise errors.UsageError(f"Cannot specify the same explicit {type} name twice.")

            return result

        explicit_inputs_parsed = parse_explicit_definition(explicit_inputs, "input")
        explicit_outputs_parsed = parse_explicit_definition(explicit_outputs, "output")
        explicit_parameters_parsed = parse_explicit_definition(explicit_parameters, "param")

        factory = PlanFactory(
            command_line=command_line,
            explicit_inputs=explicit_inputs_parsed,
            explicit_outputs=explicit_outputs_parsed,
            explicit_parameters=explicit_parameters_parsed,
            directory=os.getcwd(),
            working_dir=working_dir,
            no_input_detection=no_input_detection,
            no_output_detection=no_output_detection,
            no_parameter_detection=no_parameter_detection,
            success_codes=success_codes,
            **{name: os.path.relpath(path, working_dir) for name, path in mapped_std.items()},
        )
        with factory.watch(no_output=no_output) as tool:
            # Don't compute paths if storage is disabled.
            if check_external_storage():
                # Make sure all inputs are pulled from a storage.
                paths_ = (path for _, path in tool.iter_input_files(project_context.path))
                pull_paths_from_storage(project_context.repository, *paths_)

            if tty_exists:
                # apply original output redirection
                if stdout_redirected:
                    sys.stdout = old_stdout
                if stderr_redirected:
                    sys.stderr = old_stderr

                if "stdout" not in mapped_std:
                    try:
                        sys.stdout.fileno()
                    except UnsupportedOperation:
                        # Pytest capsys creates a pseudo device that doesn't have fileno and would fail if passed
                        pass
                    else:
                        mapped_std["stdout"] = sys.stdout

                if "stderr" not in mapped_std:
                    try:
                        sys.stderr.fileno()
                    except UnsupportedOperation:
                        # Pytest capsys creates a pseudo device that doesn't have fileno and would fail if passed
                        pass
                    else:
                        mapped_std["stderr"] = sys.stderr

            started_at_time = local_now()

            try:
                return_code = call(
                    factory.command_line, cwd=os.getcwd(), **{key: getattr(sys, key) for key in mapped_std.keys()}
                )
            except FileNotFoundError:
                command = " ".join(factory.base_command)
                raise errors.ParameterError(
                    f"Cannot execute command '{command}': "
                    "This is likely because the executable doesn't exist or has the wrong permissions set."
                )

            ended_at_time = local_now()

            sys.stdout.flush()
            sys.stderr.flush()

            if tty_exists:
                # change back to /dev/tty redirection
                if stdout_redirected:
                    sys.stdout = system_stdout  # type:ignore
                if stderr_redirected:
                    sys.stderr = system_stderr  # type:ignore

            if return_code not in (success_codes or {0}):
                raise errors.InvalidSuccessCode(return_code, success_codes=success_codes)

        if not creators:
            creators = [cast(Person, get_git_user(project_context.repository))]

        plan = tool.to_plan(
            name=name, description=description, keywords=keyword, creators=creators, date_created=started_at_time
        )
        activity = Activity.from_plan(
            plan=plan,
            repository=project_context.repository,
            started_at_time=started_at_time,
            ended_at_time=ended_at_time,
            annotations=tool.annotations,
        )
        activity_gateway.add(activity)

        return PlanViewModel.from_plan(plan)

    finally:
        if system_stdout:
            system_stdout.flush()
            sys.stdout = old_stdout
            system_stdout.close()
        if system_stderr:
            system_stderr.flush()
            sys.stderr = old_stderr
            system_stderr.close()
